
#include "Rng.h"

/****************************************************************************
 * DSAGenX()                                                                *
 *                                                                          *
 * This function generates a DSA private key X using the algorithm          *
 * given in Appendix 3.1.                                                   *
 *                                                                          *
 * INPUTS:                                                                  *
 *      XKey:  Points to an b-bit seed-key value.                           *
 *                                                                          *
 *      XSeed:  Points to an b-bit optional user selected seed value.       *
 *                                                                          *
 *      Q:  Points to DSA Q parameter used to bound value of X by alg.      *
 *                                                                          *
 *      rndtype:  Chooses what RNG method is used.                          *
 *                RAN_SHS_CN - G_from_SHA, using change notice method       *
 *                RAN_DES_CN - G_from_DES, using change notice method       *
 *                RAN_SHS - G_from_SHA, using original method               *
 *                RAN_DES - G_from_DES, using original method               *
 *                RAN_ANSI - use ANSI X9.17 for RNG                         *
 *                                                                          *
 * OUTPUTS:                                                                 *
 *      X: (size of Q)-bit private key X return value.                      *
 ****************************************************************************/
int DSAGenX(MP XKey, MP XSeed, int b, MP Q, MP *X, int rndtype)
{
	if ( rndtype & CN_MASK )
		return DSAGenX_CN(XKey, XSeed, b, Q, X, rndtype);
	else
		return DSAGenX_org(XKey, XSeed, b, Q, X, rndtype);
	}


/****************************************************************************
 * DSAGenX_org()                                                            *
 *                                                                          *
 * This function generates a DSA private key X using the algorithm          *
 * given in Appendix 3.1.                                                   *
 *                                                                          *
 * INPUTS:                                                                  *
 *      XKey:  Points to an b-bit seed-key value.                           *
 *                                                                          *
 *      XSeed:  Points to an b-bit optional user selected seed value.       *
 *                                                                          *
 *      Q:  Points to DSA Q parameter used to bound value of X by alg.      *
 *                                                                          *
 *      rndtype:  Chooses what RNG method is used.                          *
 *                RAN_SHS_CN - G_from_SHA, using change notice method       *
 *                RAN_DES_CN - G_from_DES, using change notice method       *
 *                RAN_SHS - G_from_SHA, using original method               *
 *                RAN_DES - G_from_DES, using original method               *
 *                RAN_ANSI - use ANSI X9.17 for RNG                         *
 *                                                                          *
 * OUTPUTS:                                                                 *
 *      X: (size of Q)-bit private key X return value.                      *
 ****************************************************************************/
int DSAGenX_org(MP XKey, MP XSeed, int b, MP Q, MP *X, int rndtype)
{
	BYTE	XVal[MAXSEEDLEN], Temp[MAXSEEDLEN+1], tx[MAXBYTEHASHLEN],
			Power2[MAXSEEDLEN+1], One[1], *G, mask;
	int		j, offset;
//FILE *fp = fopen("atest_Xorg.txt", "w");

	if ( ( b < MINSEEDLEN*8) || (b > MAXSEEDLEN*8) )
		return 1;

	/* set power2 = 2^keylen */
	bzero(Power2, MAXSEEDLEN+1);
	mask = 0x01 << (b%8);
	Power2[MAXSEEDLEN-(b+7)/8] = mask; //  MAXSEEDLEN+1-(b+7)/8-1, +1-1 cancel

	/* set One to the value of 1 */
	One[0] = 0x01;

	/* Compute xval = (Xkey + Xseed) mod 2^keylen */

	/* temp = Xkey left padded out to MAXSEEDLEN+1 bytes */
	bzero(XVal, sizeof(XVal));
	bzero(Temp, sizeof(Temp));
	bcopy(XKey.val, &Temp[sizeof(Temp)-XKey.size], XKey.size);
	/* temp = temp + Xseed */
	add(Temp, sizeof(Temp), XSeed.val, XSeed.size);

	/* temp = temp mod 2^keylen */
	Mod(Temp, sizeof(Temp), Power2, MAXSEEDLEN+1);
	/* xval = temp */
	bcopy(&Temp[sizeof(Temp)-XKey.size], XVal, XKey.size);

	/* Compute X = G(t, xval) mod q  */
	switch(Q.size) {
	case 20:
		bcopy((BYTE *)SHA1_IH, tx, SHA1_BYTEHASHLEN);
		break;
	case 28:
	case 32:
		bcopy((BYTE *)SHA256_IH, tx, SHA256_BYTEHASHLEN);
		break;
	case 48:
		bcopy((BYTE *)SHA384_IH, tx, SHA384_BYTEHASHLEN);
		break;
	case 64:
		bcopy((BYTE *)SHA512_IH, tx, SHA512_BYTEHASHLEN);
		break;
	}

	/* Select the proper G function */
	offset = b % 8;
	for ( j=0; j<offset; j++ )
		bshl(XVal, XKey.size);

//fprintBstr(fp, "XVal is ", XVal, 20);
	switch(rndtype) {
	case RAN_DES:
		G = (BYTE *)calloc(20, 1);
		desG(tx, XVal, G);
//fprintBstr(fp, "G() is ", G, 20);
		break;
	case RAN_SHS:
	default:
		shaG(tx, XVal, b, &G, Q.size*8);
		break;
	}
	Mod(G, Q.size, Q.val, Q.size);

	allocMP(X, Q.bitlen);
	bcopy(G, X->val, X->size);
	FREE(G);
//fprintBstr(fp, "X is ", X->val, 20);

	/********************************************
	 * Update XKEY:                             *
	 *   Compute XKEY = (1+XKEY+X) mod 2^keylen *
	 ********************************************/

	/* temp = Xkey */
	bzero(Temp, sizeof(Temp));
	bcopy(XKey.val, &Temp[sizeof(Temp)-XKey.size], XKey.size);

	/* temp = temp + X */
	add(Temp, sizeof(Temp), X->val, X->size); 
	/* temp = temp + 1 */
	add(Temp, sizeof(Temp), One, 1);

	/* temp = temp mod 2&keylen */
	Mod(Temp, sizeof(Temp), Power2, sizeof(Power2));
	/* Xkey = temp */
	bcopy(&Temp[sizeof(Temp)-XKey.size], XKey.val, XKey.size);
//fprintBstr(fp, "XKey is ", XKey.val, 20);
//fclose(fp);

	return 0;
}  /* END DSAGenX_org */


/****************************************************************************
 * DSAGenX_CN()                                                             *
 *                                                                          *
 * This function generates a DSA private key X using the algorithm          *
 * given in Appendix 3.1.                                                   *
 *                                                                          *
 * INPUTS:                                                                  *
 *      XKey:  Points to an b-bit seed-key value.                           *
 *                                                                          *
 *      XSeed:  Points to an b-bit optional user selected seed value.       *
 *                                                                          *
 *      Q:  Points to DSA Q parameter used to bound value of X by alg.      *
 *                                                                          *
 *      rndtype:  Chooses what RNG method is used.                          *
 *                RAN_SHS_CN - G_from_SHA, using change notice method       *
 *                RAN_DES_CN - G_from_DES, using change notice method       *
 *                RAN_SHS - G_from_SHA, using original method               *
 *                RAN_DES - G_from_DES, using original method               *
 *                RAN_ANSI - use ANSI X9.17 for RNG                         *
 *                                                                          *
 * OUTPUTS:                                                                 *
 *      X: (size of Q)-bit private key X return value.                      *
 ****************************************************************************/
int DSAGenX_CN(MP XKey, MP XSeed, int b, MP Q, MP *X, int rndtype)
{
	BYTE	XVal[MAXSEEDLEN], Temp[MAXSEEDLEN+1], tx[MAXBYTEHASHLEN],
			Power2[MAXSEEDLEN+1], One[1], *w, *G, mask;
	int		i, j, offset;
//FILE *fp = fopen("atest_Xcn.txt", "w");

	if ( ( b < MINSEEDLEN*8) || (b > MAXSEEDLEN*8) )
		return 1;

	/* set power2 = 2^keylen */
	bzero(Power2, MAXSEEDLEN+1);
	mask = 0x01 << (b%8);
	Power2[MAXSEEDLEN-(b+7)/8] = mask; //  MAXSEEDLEN+1-(b+7)/8-1, +1-1 cancel

	/* set One to the value of 1 */
	One[0] = 0x01;

	switch(Q.size) {
	case 20:
		bcopy((BYTE *)SHA1_IH, tx, SHA1_BYTEHASHLEN);
		w = (BYTE *)calloc(2, SHA1_BYTEHASHLEN);
		break;
	case 28:
	case 32:
		bcopy((BYTE *)SHA256_IH, tx, SHA256_BYTEHASHLEN);
		w = (BYTE *)calloc(2, SHA256_BYTEHASHLEN);
		break;
	case 48:
		bcopy((BYTE *)SHA384_IH, tx, SHA384_BYTEHASHLEN);
		w = (BYTE *)calloc(2, SHA384_BYTEHASHLEN);
		break;
	case 64:
		bcopy((BYTE *)SHA512_IH, tx, SHA512_BYTEHASHLEN);
		w = (BYTE *)calloc(2, SHA512_BYTEHASHLEN);
		break;
	}

	for ( i=0; i<2; i++ ) {
	
		/* Compute xval = (Xkey + Xseed) mod 2^keylen */
		bzero(XVal, sizeof(XVal));
		bzero(Temp, sizeof(Temp));
		bcopy(XKey.val, &Temp[sizeof(Temp)-XKey.size], XKey.size);
		/* temp = temp + Xseed */
		add(Temp, sizeof(Temp), XSeed.val, XSeed.size);

		/* temp = temp mod 2^keylen */
		Mod(Temp, sizeof(Temp), Power2, sizeof(Power2));
		/* xval = temp */
		bcopy(&Temp[sizeof(Temp)-XKey.size], XVal, XKey.size);

		/* Compute w[i] = G(t, xval) */

		offset = b % 8;
		for ( j=0; j<offset; j++ )
			bshl(XVal, XKey.size);

		switch(rndtype) {
		case RAN_DES_CN:
//fprintBstr(fp, "XVal is ", XVal, 20);
			desG(tx, XVal, &w[i*Q.size]);
//fprintBstr(fp, "G() is ", &w[i*Q.size], 20);
			break;
		case RAN_SHS_CN:
		default:
			G = NULL;
			shaG(tx, XVal, b, &G, Q.size*8);
			bcopy(G, w+i*Q.size, Q.size);
			FREE(G);
			break;
		}
		/***********************************************
		 * Update XKEY:                                *
		 *   Compute XKEY = (1+XKEY+w[i]) mod 2^keylen *
		 ***********************************************/

		/* temp = Xkey */
		bzero(Temp, sizeof(Temp));
		bcopy(XKey.val, &Temp[sizeof(Temp)-XKey.size], XKey.size);

		/* temp = temp + w_i */
		add(Temp, sizeof(Temp), &w[i*Q.size], Q.size); 
		/* temp = temp + 1 */
		add(Temp, sizeof(Temp), One, 1);

		/* temp = temp mod 2^keylen */
		Mod(Temp, sizeof(Temp), Power2, sizeof(Power2));
		/* Xkey = temp */
		bcopy(&Temp[sizeof(Temp)-XKey.size], XKey.val, XKey.size);
//fprintBstr(fp, "XKey is ", XKey.val, 20);

	}

	Mod(w, 2*Q.size, Q.val, Q.size);
	allocMP(X, Q.bitlen);
	X->size = Q.size;
	if ( X->val )
		free(X->val);
	X->val = (BYTE *)malloc(X->size);
	bcopy(w+Q.size, X->val, X->size);
	FREE(w);
//fprintBstr(fp, "X is ", X->val, 20);
//fclose(fp);

	return 0;
}  /* END DSAGenX_CN */


/*****************************************************************************
 * DSAGenK()                                                                 *
 *                                                                           *
 *   This function generates a DSA signature parameter K using the algorithm *
 *   given in Appendix 3.2.                                                  *
 *                                                                           *
 *      keylen:	Length b of Kkey in bits (160-512)                           *
 *              NOTE: this needs to be added for the next version            *
 *                                                                           *
 *   INPUTS:                                                                 *
 *      Kkey:   Points to an b-bit seed-key value.                           *
 *                                                                           *
 *      Q:      Points to DSA Q parameter used to bound value of K by alg.   *
 *                                                                           *
 *      rndtype:  Chooses what RNG method is used.                           *
 *                RAN_SHS_CN - G_from_SHA, using change notice method        *
 *                RAN_DES_CN - G_from_DES, using change notice method        *
 *                RAN_SHS - G_from_SHA, using original method                *
 *                RAN_DES - G_from_DES, using original method                *
 *                RAN_ANSI - use ANSI X9.17 for RNG                          *
 *   OUTPUTS:                                                                *
 *      K:  160-bit signature parameter K return value.                      *
 *****************************************************************************/
int DSAGenK(MP KKey, int b, MP Q, MP *K, int rndtype)
{
	if ( rndtype & CN_MASK )
		return DSAGenK_CN(KKey, b, Q, K, rndtype);
	else
		return DSAGenK_org(KKey, b, Q, K, rndtype);

	return 0;
	}


/*****************************************************************************
 * DSAGenK_org()                                                             *
 *                                                                           *
 *   This function generates a DSA signature parameter K using the algorithm *
 *   given in Appendix 3.2.                                                  *
 *                                                                           *
 *      keylen:	Length b of Kkey in bits (160-512)                           *
 *              NOTE: this needs to be added for the next version            *
 *                                                                           *
 *   INPUTS:                                                                 *
 *      Kkey:   Points to an b-bit seed-key value.                           *
 *                                                                           *
 *      Q:      Points to DSA Q parameter used to bound value of K by alg.   *
 *                                                                           *
 *      rndtype:  Chooses what RNG method is used.                           *
 *                RAN_SHS_CN - G_from_SHA, using change notice method        *
 *                RAN_DES_CN - G_from_DES, using change notice method        *
 *                RAN_SHS - G_from_SHA, using original method                *
 *                RAN_DES - G_from_DES, using original method                *
 *                RAN_ANSI - use ANSI X9.17 for RNG                          *
 *   OUTPUTS:                                                                *
 *      K:  160-bit signature parameter K return value.                      *
 *****************************************************************************/
int DSAGenK_org(MP KKey, int b, MP Q, MP *K, int rndtype)
{
	BYTE	One[1], Temp[MAXSEEDLEN+1], Power2[MAXSEEDLEN+1], tk[MAXBYTEHASHLEN],
			*G, mask;
	int		j, offset;
//FILE *fp = fopen("atest_Korg.txt", "w");

	if ( ( b < MINSEEDLEN*8) || (b > MAXSEEDLEN*8) )
		return 1;

	/* set power2 = 2^keylen */
	bzero(Power2, MAXSEEDLEN+1);
	mask = 0x01 << (b%8);
	Power2[MAXSEEDLEN-(b+7)/8] = mask; //  MAXSEEDLEN+1-(b+7)/8-1, +1-1 cancel

	/* set One to the value of 1 */
	One[0] = 0x01;

	switch(Q.size) {
	case 20:
		bcopy((BYTE *)SHA1_IH, tk, SHA1_BYTEHASHLEN);
		ROTL_ULONGS((ULONG *)tk, SHA1_ULONGHASHLEN);
		break;
	case 28:
	case 32:
		bcopy((BYTE *)SHA256_IH, tk, SHA256_BYTEHASHLEN);
		ROTL_ULONGS((ULONG *)tk, SHA256_ULONGHASHLEN);
		break;
	case 48:
		bcopy((BYTE *)SHA384_IH, tk, SHA384_BYTEHASHLEN);
		ROTL_ULONGS((ULONG *)tk, SHA384_ULONGHASHLEN);
		break;
	case 64:
		bcopy((BYTE *)SHA512_IH, tk, SHA512_BYTEHASHLEN);
		ROTL_ULONGS((ULONG *)tk, SHA512_ULONGHASHLEN);
		break;
	}

	/* Compute K = G(t, KKEY) mod q  */
	offset = b % 8;
	for ( j=0; j<offset; j++ )
		bshl(KKey.val, KKey.size);

//fprintBstr(fp, "KKey is ", KKey.val, 20);
	switch(rndtype) {
	case RAN_DES:
		G = (BYTE *)calloc(20, 1);
		desG(tk, KKey.val, G);
		break;
	case RAN_SHS:
	default:
		shaG(tk, KKey.val, b, &G, Q.size*8);
		break;
	}
//fprintBstr(fp, "G() is ", G, 20);

	Mod(G, Q.size, Q.val, Q.size);
	allocMP(K, Q.bitlen);
	bcopy(G, K->val, K->size);
	FREE(G);
//fprintBstr(fp, "K is ", K->val, 20);

	/**********************************************
	 * Update KKEY:                               *
	 *   Compute KKEY = (1+KKEY+K) mod 2^keylen   *
	 **********************************************/

	/* temp = Kkey */
	bzero(Temp, sizeof(Temp));
	bcopy(KKey.val, &Temp[sizeof(Temp)-KKey.size], KKey.size);

	/* temp = temp + K */
	add(Temp, sizeof(Temp), K->val, K->size);
	/* temp = temp + 1 */
	add(Temp, sizeof(Temp), One, 1);

	/* temp = temp mod 2&keylen */
	Mod(Temp, sizeof(Temp), Power2, sizeof(Power2));

	/* Kkey = temp */
	bcopy(&Temp[sizeof(Temp)-KKey.size], KKey.val, KKey.size);
//fprintBstr(fp, "KKey is ", KKey.val, 20);
//fclose(fp);

	return 0;
}  /* END DSAGenK_org */


/*****************************************************************************
 * DSAGenK_CN()                                                              *
 *                                                                           *
 *   This function generates a DSA signature parameter K using the algorithm *
 *   given in Appendix 3.2.                                                  *
 *                                                                           *
 *      keylen:	Length b of Kkey in bits (160-512)                           *
 *              NOTE: this needs to be added for the next version            *
 *                                                                           *
 *   INPUTS:                                                                 *
 *      Kkey:   Points to an b-bit seed-key value.                           *
 *                                                                           *
 *      Q:      Points to DSA Q parameter used to bound value of K by alg.   *
 *                                                                           *
 *      rndtype:  Chooses what RNG method is used.                           *
 *                RAN_SHS_CN - G_from_SHA, using change notice method        *
 *                RAN_DES_CN - G_from_DES, using change notice method        *
 *                RAN_SHS - G_from_SHA, using original method                *
 *                RAN_DES - G_from_DES, using original method                *
 *                RAN_ANSI - use ANSI X9.17 for RNG                          *
 *   OUTPUTS:                                                                *
 *      K:  160-bit signature parameter K return value.                      *
 *****************************************************************************/
int DSAGenK_CN(MP KKey, int b, MP Q, MP *K, int rndtype)
{
	BYTE	Temp[MAXSEEDLEN+1], tk[MAXBYTEHASHLEN], Power2[MAXSEEDLEN+1], One[1],
			*w, *G, mask;
	int		i, j, offset;
//FILE *fp = fopen("atest_Kcn.txt", "w");

	if ( ( b < MINSEEDLEN*8) || (b > MAXSEEDLEN*8) )
		return 1;

	/* set power2 = 2^keylen */
	bzero(Power2, MAXSEEDLEN+1);
	mask = 0x01 << (b%8);
	Power2[MAXSEEDLEN-(b+7)/8] = mask; //  MAXSEEDLEN+1-(b+7)/8-1, +1-1 cancel

	/* set One to the value of 1 */
	One[0] = 0x01;

	switch(Q.size) {
	case 20:
		bcopy((BYTE *)SHA1_IH, tk, SHA1_BYTEHASHLEN);
		ROTL_ULONGS((ULONG *)tk, SHA1_ULONGHASHLEN);
		w = (BYTE *)calloc(2, SHA1_BYTEHASHLEN);
		break;
	case 28:
	case 32:
		bcopy((BYTE *)SHA256_IH, tk, SHA256_BYTEHASHLEN);
		ROTL_ULONGS((ULONG *)tk, SHA256_ULONGHASHLEN);
		w = (BYTE *)calloc(2, SHA256_BYTEHASHLEN);
		break;
	case 48:
		bcopy((BYTE *)SHA384_IH, tk, SHA384_BYTEHASHLEN);
		ROTL_ULONGS((ULONG *)tk, SHA384_ULONGHASHLEN);
		w = (BYTE *)calloc(2, SHA384_BYTEHASHLEN);
		break;
	case 64:
		bcopy((BYTE *)SHA512_IH, tk, SHA512_BYTEHASHLEN);
		ROTL_ULONGS((ULONG *)tk, SHA512_ULONGHASHLEN);
		w = (BYTE *)calloc(2, SHA512_BYTEHASHLEN);
		break;
	}

//fprintBstr(fp, "KKey is ", KKey.val, KKey.size);
	for ( i=0; i<2; i++ ) {
	
		/* Compute w[i] = G(t, xval) */
		offset = b % 8;
		for ( j=0; j<offset; j++ )
			bshl(KKey.val, KKey.size);

		switch(rndtype) {
		case RAN_DES_CN:
			desG(tk, KKey.val, &w[i*Q.size]);
//fprintBstr(fp, "G() is ", &w[i*Q.size], Q.size);
			break;
		case RAN_SHS_CN:
		default:
			G = NULL;
			shaG(tk, KKey.val, b, &G, Q.size*8);
			bcopy(G, &w[i*Q.size], Q.size);
			FREE(G);
			break;
		}
		/***********************************************
		* Update KKEY:                                *
		*   Compute KKEY = (1+KKEY+w[i]) mod 2^keylen *
		***********************************************/

		/* temp = KKey */
		bzero(Temp, sizeof(Temp));
		bcopy(KKey.val, &Temp[sizeof(Temp)-KKey.size], KKey.size);

		/* temp = temp + w_i */
		add(Temp, sizeof(Temp), &w[i*Q.size], Q.size);
		/* temp = temp + 1 */
		add(Temp, sizeof(Temp), One, 1);

		/* temp = temp mod 2^keylen */
		Mod(Temp, sizeof(Temp), Power2, sizeof(Power2));
		/* KKey = temp */
		bcopy(&Temp[sizeof(Temp)-KKey.size], KKey.val, KKey.size);
//fprintBstr(fp, "KKey is ", KKey.val, KKey.size);
	}

	Mod(w, 2*Q.size, Q.val, Q.size);
	allocMP(K, Q.bitlen);
	K->size = Q.size;
	if ( K->val )
		free(K->val);
	K->val = (BYTE *)malloc(K->size);
	bcopy(w+Q.size, K->val, K->size);
	FREE(w);
//fprintBstr(fp, "K is ", K->val, 20);
//fclose(fp);

	return 0;
}  /* END DSAGenK_CN */


/****************************************************************************
 * shaG()                                                                   *
 *                                                                          *
 *  This is the implementation of the G-function based on a subset          *
 *  of SHA steps as defined in Appendix 3.3.                                *
 *                                                                          *
 *  INPUTS:                                                                 *
 *      tval:    Points to initial hash array (IH) to initialize SHA        *
 *                                                                          *
 *      b:       Length of c in bits <= 512                                 *
 *                                                                          *
 *      C:       160-512 bit SEED.                                          *
 *                                                                          *
 *      hashlen: selects the hash algorithm to use.                         *
 *                                                                          *
 *  OUTPUTS:                                                                *
 *      G:  160-bit return value                                            *
 ****************************************************************************/
void shaG(BYTE *tval, BYTE *C, int b, BYTE **G, int hashlen)
{
	BYTE	*M1;
	HASH	ctx;
	int		i, CBytes;

	ctx.hashlen = hashlen;
	CBytes = (b+7)/8;
	switch (hashlen) {
	case 160:
		M1 = (BYTE *)calloc(SHA1_BYTEBLOCKLEN, 1);
		bcopy(C, M1, CBytes);
		SHA1_Init2 (&(ctx.u_ctx.ctx1), (ULONG *)tval);
		SHA1_Update (&(ctx.u_ctx.ctx1), M1, SHA1_BLOCKLEN);
		*G = (BYTE *)malloc(SHA1_BYTEHASHLEN);
		bcopy((BYTE *)ctx.u_ctx.ctx1.H, *G, SHA1_BYTEHASHLEN);
#ifdef LITTLE_ENDIAN
		c2LH(*G, SHA1_BYTEHASHLEN);
		for ( i=0; i<SHA1_ULONGHASHLEN; i++ )
			cLONG2DIGITS((DIGIT *)&((*G)[i*4]));
#endif
		break;
	case 224:
	case 256:
		M1 = (BYTE *)calloc(SHA256_BYTEBLOCKLEN, 1);
		bcopy(C, M1, CBytes);
		SHA256_Init2 (&(ctx.u_ctx.ctx256), (ULONG *)tval);
		SHA256_Update (&(ctx.u_ctx.ctx256), M1, SHA256_BLOCKLEN);
		*G = (BYTE *)malloc(SHA256_BYTEHASHLEN);
		bcopy((BYTE *)ctx.u_ctx.ctx256.H, *G, SHA256_BYTEHASHLEN);
#ifdef LITTLE_ENDIAN
		c2LH(*G, SHA256_BYTEHASHLEN);
		for ( i=0; i<SHA256_ULONGHASHLEN; i++ )
			cLONG2DIGITS((DIGIT *)&((*G)[i*4]));
#endif
		break;
	case 384:
		M1 = (BYTE *)calloc(SHA384_BYTEBLOCKLEN, 1);
		bcopy(C, M1, CBytes);
		SHA384_Init2 (&(ctx.u_ctx.ctx384), (WORD64 *)tval);
		SHA384_Update (&(ctx.u_ctx.ctx384), M1, SHA384_BLOCKLEN);
		*G = (BYTE *)malloc(SHA384_BYTEHASHLEN);
		bcopy((BYTE *)ctx.u_ctx.ctx384.H, *G, SHA384_BYTEHASHLEN);
#ifdef LITTLE_ENDIAN
		c2LH(*G, SHA384_BYTEHASHLEN);
		for ( i=0; i<SHA384_ULONGHASHLEN; i++ )
			cLONG2DIGITS((DIGIT *)&((*G)[i*4]));
#endif
		break;
	case 512:
		M1 = (BYTE *)calloc(SHA512_BYTEBLOCKLEN, 1);
		bcopy(C, M1, CBytes);
		SHA512_Init2 (&(ctx.u_ctx.ctx512), (WORD64 *)tval);
		SHA512_Update (&(ctx.u_ctx.ctx512), M1, SHA512_BLOCKLEN);
		*G = (BYTE *)malloc(SHA512_BYTEHASHLEN);
		bcopy((BYTE *)ctx.u_ctx.ctx256.H, *G, SHA512_BYTEHASHLEN);
#ifdef LITTLE_ENDIAN
		c2LH(*G, SHA512_BYTEHASHLEN);
		for ( i=0; i<SHA512_ULONGHASHLEN; i++ )
			cLONG2DIGITS((DIGIT *)&((*G)[i*4]));
#endif
		break;
	}

	free(M1);
}

/*********************************************************
 * desG:                                                 *
 *        NOTE: t and c are assumed to be 160 bits long  *
 *********************************************************/
void desG(BYTE *t, BYTE *c, BYTE *G)
{
	BYTE	x[20], y[40], a1[4], a2[4], b1[4], b2[4], tmp[4];
	BYTE	A[8];
	BYTE	K56[7], K64[8];
	int		i;
	TDESCTX context;

#ifdef LITTLE_ENDIAN
	c2LH(t, SHA1_BYTEHASHLEN);
	for ( i=0; i<SHA1_ULONGHASHLEN; i++ )
		cLONG2DIGITS((DIGIT *)&((t)[i*4]));
#endif

	for ( i=0; i<20; i++ )		/* Step 2 of the Appendix */
		x[i] = t[i] ^ c[i];

	for ( i=0; i<5; i++ )  {	/* Step 3 of the Appendix */
		bcopy(c+((i+4)%5)*4, b1, 4);
		bcopy(c+((i+3)%5)*4, b2, 4);
		bcopy(x+i*4, a1, 4);
		xor(a2, x+((i+1)%5)*4, x+((i+4)%5)*4, 4);
		bcopy(a1, A, 4);
		bcopy(a2, A+4, 4);

		bcopy(b1+1, K56, 3);
		bcopy(b2, K56+3, 4);
		key56to64(K56, K64);

		//K[0] = 0x10; K[1] = 0x31; K[2] = 0x6e; K[3] = 0x02;
		//K[4] = 0x8c; K[5] = 0x8f; K[6] = 0x3b; K[7] = 0x4a;
		TDESInit (&context, K64, K64, K64, NULL, TDES_ECB, DES_ENCIPHER);

		//setkey(IGNOREPARITY, ENCRYPT, K, ks);

		//A[0] = 0x00; A[1] = 0x00; A[2] = 0x00; A[3] = 0x00;
		//A[4] = 0x00; A[5] = 0x00; A[6] = 0x00; A[7] = 0x00;
		TDES(&context, A, A, NULL);
		//des(A, A, ks);				/* Replace A with encrypted A */

		bcopy(A, &(y[i*8]), 8);
	}

	for (i=0; i<5; i++) {			/* Complete step 4 and store */
		xor(tmp, &(y[i*8]), &(y[((i+2)%5)*8+4]), 4);
		xor(G+i*4, tmp, &(y[((i+3)%5)*8]), 4);
	}
#ifdef LITTLE_ENDIAN
	c2LH(t, SHA1_BYTEHASHLEN);
	for ( i=0; i<SHA1_ULONGHASHLEN; i++ )
		cLONG2DIGITS((DIGIT *)&((t)[i*4]));
#endif
}

int DSARngTest()
{
	MP	Q, XKey, XSeed, KKey, X, K;
//	FILE	*fp;
//	fp = fopen("atest2.txt", "w");

	initMP(&Q);
	initMP(&XKey);
	initMP(&XSeed);
	initMP(&KKey);
	initMP(&X);
	initMP(&K);

	allocMP(&Q, 160);

	Q.val[0] = 0xb5; Q.val[1] = 0xaf;
	Q.val[2] = 0xd2; Q.val[3] = 0xf9;
	Q.val[4] = 0x32; Q.val[5] = 0x46;
	Q.val[6] = 0xb1; Q.val[7] = 0xef;
	Q.val[8] = 0xcd; Q.val[9] = 0x1f;
	Q.val[10] = 0x3a; Q.val[11] = 0x7c;
	Q.val[12] = 0x24; Q.val[13] = 0x0c;
	Q.val[14] = 0x1e; Q.val[15] = 0x9e;
	Q.val[16] = 0x21; Q.val[17] = 0xa3;
	Q.val[18] = 0x63; Q.val[19] = 0x0b;

	allocMP(&XKey, 160);
	allocMP(&XSeed, 160);
	allocMP(&KKey, 160);
	
	/*  Test FIPS 186 Gen X - Original w/ SHA-Gfnc  */
	XKey.val[0] = 0xbd; XKey.val[1] = 0x02;
	XKey.val[2] = 0x9b; XKey.val[3] = 0xbe;
	XKey.val[4] = 0x7f; XKey.val[5] = 0x51;
	XKey.val[6] = 0x96; XKey.val[7] = 0x0b;
	XKey.val[8] = 0xcf; XKey.val[9] = 0x9e;
	XKey.val[10] = 0xdb; XKey.val[11] = 0x2b;
	XKey.val[12] = 0x61; XKey.val[13] = 0xf0;
	XKey.val[14] = 0x6f; XKey.val[15] = 0x0f;
	XKey.val[16] = 0xeb; XKey.val[17] = 0x5a;
	XKey.val[18] = 0x38; XKey.val[19] = 0xb6;
	bzero(XSeed.val, XSeed.size);

	DSAGenX(XKey, XSeed, 160, Q, &X, RAN_SHS);
	
	if ( (X.val[0] != 0x20) || (X.val[1] != 0x70) ||
		 (X.val[2] != 0xb3) || (X.val[3] != 0x22) ||
		 (X.val[4] != 0x3d) || (X.val[5] != 0xba) ||
		 (X.val[6] != 0x37) || (X.val[7] != 0x2f) ||
		 (X.val[8] != 0xde) || (X.val[9] != 0x1c) ||
		 (X.val[10] != 0x0f) || (X.val[11] != 0xfc) ||
		 (X.val[12] != 0x7b) || (X.val[13] != 0x2e) ||
		 (X.val[14] != 0x3b) || (X.val[15] != 0x49) ||
		 (X.val[16] != 0x8b) || (X.val[17] != 0x26) ||
		 (X.val[18] != 0x06) || (X.val[19] != 0x14) )
		return 1;
	
	freeMP(&X);

	/*  Test FIPS 186 Gen X - Change Notice w/ SHA-Gfnc  */
	XKey.val[0] = 0xbd; XKey.val[1] = 0x02;
	XKey.val[2] = 0x9b; XKey.val[3] = 0xbe;
	XKey.val[4] = 0x7f; XKey.val[5] = 0x51;
	XKey.val[6] = 0x96; XKey.val[7] = 0x0b;
	XKey.val[8] = 0xcf; XKey.val[9] = 0x9e;
	XKey.val[10] = 0xdb; XKey.val[11] = 0x2b;
	XKey.val[12] = 0x61; XKey.val[13] = 0xf0;
	XKey.val[14] = 0x6f; XKey.val[15] = 0x0f;
	XKey.val[16] = 0xeb; XKey.val[17] = 0x5a;
	XKey.val[18] = 0x38; XKey.val[19] = 0xb6;
	bzero(XSeed.val, XSeed.size);

	DSAGenX(XKey, XSeed, 160, Q, &X, RAN_SHS_CN);
	
	if ( (X.val[0] != 0x47) || (X.val[1] != 0xc2) ||
		 (X.val[2] != 0x7e) || (X.val[3] != 0xb6) ||
		 (X.val[4] != 0x16) || (X.val[5] != 0xdb) ||
		 (X.val[6] != 0xa4) || (X.val[7] != 0x13) ||
		 (X.val[8] != 0x91) || (X.val[9] != 0xe5) ||
		 (X.val[10] != 0x16) || (X.val[11] != 0x5b) ||
		 (X.val[12] != 0xe9) || (X.val[13] != 0xc5) ||
		 (X.val[14] != 0xe3) || (X.val[15] != 0x97) ||
		 (X.val[16] != 0x7e) || (X.val[17] != 0x39) ||
		 (X.val[18] != 0xa1) || (X.val[19] != 0x5d) )
		return 2;

	freeMP(&X);

	/*  Test FIPS 186 Gen K - Original w/ SHA-Gfnc  */
	KKey.val[0] = 0x68; KKey.val[1] = 0x7a;
	KKey.val[2] = 0x66; KKey.val[3] = 0xd9;
	KKey.val[4] = 0x06; KKey.val[5] = 0x48;
	KKey.val[6] = 0xf9; KKey.val[7] = 0x93;
	KKey.val[8] = 0x86; KKey.val[9] = 0x7e;
	KKey.val[10] = 0x12; KKey.val[11] = 0x1f;
	KKey.val[12] = 0x4d; KKey.val[13] = 0xdf;
	KKey.val[14] = 0x9d; KKey.val[15] = 0xdb;
	KKey.val[16] = 0x01; KKey.val[17] = 0x20;
	KKey.val[18] = 0x55; KKey.val[19] = 0x84;

	DSAGenK(KKey, 160, Q, &K, RAN_SHS);
	
	if ( (K.val[0] != 0x47) || (K.val[1] != 0x50) ||
		 (K.val[2] != 0xfb) || (K.val[3] != 0xea) ||
		 (K.val[4] != 0x55) || (K.val[5] != 0x9a) ||
		 (K.val[6] != 0x88) || (K.val[7] != 0x0e) ||
		 (K.val[8] != 0x1c) || (K.val[9] != 0xfe) ||
		 (K.val[10] != 0x69) || (K.val[11] != 0x80) ||
		 (K.val[12] != 0xe3) || (K.val[13] != 0x5c) ||
		 (K.val[14] != 0x41) || (K.val[15] != 0x1c) ||
		 (K.val[16] != 0x97) || (K.val[17] != 0xd6) ||
		 (K.val[18] != 0x0e) || (K.val[19] != 0x13) )
		return 3;
	
	freeMP(&K);

	/*  Test FIPS 186 Gen K - Change Notice w/ SHA-Gfnc  */
	KKey.val[0] = 0x68; KKey.val[1] = 0x7a;
	KKey.val[2] = 0x66; KKey.val[3] = 0xd9;
	KKey.val[4] = 0x06; KKey.val[5] = 0x48;
	KKey.val[6] = 0xf9; KKey.val[7] = 0x93;
	KKey.val[8] = 0x86; KKey.val[9] = 0x7e;
	KKey.val[10] = 0x12; KKey.val[11] = 0x1f;
	KKey.val[12] = 0x4d; KKey.val[13] = 0xdf;
	KKey.val[14] = 0x9d; KKey.val[15] = 0xdb;
	KKey.val[16] = 0x01; KKey.val[17] = 0x20;
	KKey.val[18] = 0x55; KKey.val[19] = 0x84;

	DSAGenK(KKey, 160, Q, &K, RAN_SHS_CN);
	
	if ( (K.val[0] != 0x95) || (K.val[1] != 0x21) ||
		 (K.val[2] != 0x27) || (K.val[3] != 0xc8) ||
		 (K.val[4] != 0xc4) || (K.val[5] != 0xb3) ||
		 (K.val[6] != 0x8b) || (K.val[7] != 0x8b) ||
		 (K.val[8] != 0xff) || (K.val[9] != 0xb0) ||
		 (K.val[10] != 0xde) || (K.val[11] != 0xfa) ||
		 (K.val[12] != 0x5f) || (K.val[13] != 0xf6) ||
		 (K.val[14] != 0xaf) || (K.val[15] != 0x91) ||
		 (K.val[16] != 0xa2) || (K.val[17] != 0xa8) ||
		 (K.val[18] != 0x12) || (K.val[19] != 0x96) )
		return 4;
	
	freeMP(&K);

	/*  Test FIPS 186 Gen X - Original w/ DES-Gfnc  */
	XKey.val[0] = 0xbd; XKey.val[1] = 0x02;
	XKey.val[2] = 0x9b; XKey.val[3] = 0xbe;
	XKey.val[4] = 0x7f; XKey.val[5] = 0x51;
	XKey.val[6] = 0x96; XKey.val[7] = 0x0b;
	XKey.val[8] = 0xcf; XKey.val[9] = 0x9e;
	XKey.val[10] = 0xdb; XKey.val[11] = 0x2b;
	XKey.val[12] = 0x61; XKey.val[13] = 0xf0;
	XKey.val[14] = 0x6f; XKey.val[15] = 0x0f;
	XKey.val[16] = 0xeb; XKey.val[17] = 0x5a;
	XKey.val[18] = 0x38; XKey.val[19] = 0xb6;
	bzero(XSeed.val, XSeed.size);

	DSAGenX(XKey, XSeed, 160, Q, &X, RAN_DES);
	
	// 3462 026c c546 cd8c 1d3b 54d1 d9f7 1860 efc2 ee63
	if ( (X.val[0] != 0x34) || (X.val[1] != 0x62) ||
		 (X.val[2] != 0x02) || (X.val[3] != 0x6c) ||
		 (X.val[4] != 0xc5) || (X.val[5] != 0x46) ||
		 (X.val[6] != 0xcd) || (X.val[7] != 0x8c) ||
		 (X.val[8] != 0x1d) || (X.val[9] != 0x3b) ||
		 (X.val[10] != 0x54) || (X.val[11] != 0xd1) ||
		 (X.val[12] != 0xd9) || (X.val[13] != 0xf7) ||
		 (X.val[14] != 0x18) || (X.val[15] != 0x60) ||
		 (X.val[16] != 0xef) || (X.val[17] != 0xc2) ||
		 (X.val[18] != 0xee) || (X.val[19] != 0x63) )
		return 5;
	
	freeMP(&X);

	/*  Test FIPS 186 Gen X - Change Notice w/ DES-Gfnc  */
	XKey.val[0] = 0xbd; XKey.val[1] = 0x02;
	XKey.val[2] = 0x9b; XKey.val[3] = 0xbe;
	XKey.val[4] = 0x7f; XKey.val[5] = 0x51;
	XKey.val[6] = 0x96; XKey.val[7] = 0x0b;
	XKey.val[8] = 0xcf; XKey.val[9] = 0x9e;
	XKey.val[10] = 0xdb; XKey.val[11] = 0x2b;
	XKey.val[12] = 0x61; XKey.val[13] = 0xf0;
	XKey.val[14] = 0x6f; XKey.val[15] = 0x0f;
	XKey.val[16] = 0xeb; XKey.val[17] = 0x5a;
	XKey.val[18] = 0x38; XKey.val[19] = 0xb6;
	bzero(XSeed.val, XSeed.size);

	DSAGenX(XKey, XSeed, 160, Q, &X, RAN_DES_CN);
	
	// 2e10 8577 08b9 fde1 2af8 30d1 2a02 8ff6 df7c 5c8f
	if ( (X.val[0] != 0x2e) || (X.val[1] != 0x10) ||
		 (X.val[2] != 0x85) || (X.val[3] != 0x77) ||
		 (X.val[4] != 0x08) || (X.val[5] != 0xb9) ||
		 (X.val[6] != 0xfd) || (X.val[7] != 0xe1) ||
		 (X.val[8] != 0x2a) || (X.val[9] != 0xf8) ||
		 (X.val[10] != 0x30) || (X.val[11] != 0xd1) ||
		 (X.val[12] != 0x2a) || (X.val[13] != 0x02) ||
		 (X.val[14] != 0x8f) || (X.val[15] != 0xf6) ||
		 (X.val[16] != 0xdf) || (X.val[17] != 0x7c) ||
		 (X.val[18] != 0x5c) || (X.val[19] != 0x8f) )
		return 6;

	freeMP(&X);

	/*  Test FIPS 186 Gen K - Original w/ DES-Gfnc  */
	KKey.val[0] = 0x68; KKey.val[1] = 0x7a;
	KKey.val[2] = 0x66; KKey.val[3] = 0xd9;
	KKey.val[4] = 0x06; KKey.val[5] = 0x48;
	KKey.val[6] = 0xf9; KKey.val[7] = 0x93;
	KKey.val[8] = 0x86; KKey.val[9] = 0x7e;
	KKey.val[10] = 0x12; KKey.val[11] = 0x1f;
	KKey.val[12] = 0x4d; KKey.val[13] = 0xdf;
	KKey.val[14] = 0x9d; KKey.val[15] = 0xdb;
	KKey.val[16] = 0x01; KKey.val[17] = 0x20;
	KKey.val[18] = 0x55; KKey.val[19] = 0x84;

	DSAGenK(KKey, 160, Q, &K, RAN_DES);
	
	// 4852 0cee 821b c35e 1cc5 acef df04 cec2 38b5 a952
	if ( (K.val[0] != 0x48) || (K.val[1] != 0x52) ||
		 (K.val[2] != 0x0c) || (K.val[3] != 0xee) ||
		 (K.val[4] != 0x82) || (K.val[5] != 0x1b) ||
		 (K.val[6] != 0xc3) || (K.val[7] != 0x5e) ||
		 (K.val[8] != 0x1c) || (K.val[9] != 0xc5) ||
		 (K.val[10] != 0xac) || (K.val[11] != 0xef) ||
		 (K.val[12] != 0xdf) || (K.val[13] != 0x04) ||
		 (K.val[14] != 0xce) || (K.val[15] != 0xc2) ||
		 (K.val[16] != 0x38) || (K.val[17] != 0xb5) ||
		 (K.val[18] != 0xa9) || (K.val[19] != 0x52) )
		return 7;
	
	freeMP(&K);

	/*  Test FIPS 186 Gen K - Change Notice w/ DES-Gfnc  */
	KKey.val[0] = 0x68; KKey.val[1] = 0x7a;
	KKey.val[2] = 0x66; KKey.val[3] = 0xd9;
	KKey.val[4] = 0x06; KKey.val[5] = 0x48;
	KKey.val[6] = 0xf9; KKey.val[7] = 0x93;
	KKey.val[8] = 0x86; KKey.val[9] = 0x7e;
	KKey.val[10] = 0x12; KKey.val[11] = 0x1f;
	KKey.val[12] = 0x4d; KKey.val[13] = 0xdf;
	KKey.val[14] = 0x9d; KKey.val[15] = 0xdb;
	KKey.val[16] = 0x01; KKey.val[17] = 0x20;
	KKey.val[18] = 0x55; KKey.val[19] = 0x84;

	DSAGenK(KKey, 160, Q, &K, RAN_DES_CN);
	
	// ab2a b897 ce90 c05a 343c c115 4afa 19a5 0170 d7ef
	if ( (K.val[0] != 0xab) || (K.val[1] != 0x2a) ||
		 (K.val[2] != 0xb8) || (K.val[3] != 0x97) ||
		 (K.val[4] != 0xce) || (K.val[5] != 0x90) ||
		 (K.val[6] != 0xc0) || (K.val[7] != 0x5a) ||
		 (K.val[8] != 0x34) || (K.val[9] != 0x3c) ||
		 (K.val[10] != 0xc1) || (K.val[11] != 0x15) ||
		 (K.val[12] != 0x4a) || (K.val[13] != 0xfa) ||
		 (K.val[14] != 0x19) || (K.val[15] != 0xa5) ||
		 (K.val[16] != 0x01) || (K.val[17] != 0x70) ||
		 (K.val[18] != 0xd7) || (K.val[19] != 0xef) )
		return 8;

	freeMP(&Q);
	freeMP(&XKey);
	freeMP(&XSeed);
	freeMP(&KKey);
	freeMP(&X);
	freeMP(&K);
//fclose(fp);

	return 0;
}

/***********************/
/*  ANSI X9.62 - ECDSA */
/***********************/
void ECDSAGenX(MP XKey, MP XSeed, int b, MP N, MP *X, int rndtype)
{
	BYTE	*XVal, *Temp, *One, *w, *G, tx[SHA1_BYTEHASHLEN];
	MP		Power2;
	int		i, j, s, f, offset;

	initMP(&Power2);
	One = Temp = w = XVal = NULL;

	s = getbitsize(N.val, N.bitlen);
	f = (s+159)/SHA1_HASHLEN;

	/* set power2 = 2^keylen */
	allocMP(&Power2, b+1);
	bzero(Power2.val, Power2.size);
	Power2.val[Power2.size-1] = 0x01;

	for ( i=0; i<b; i++ )
		bshl(Power2.val, Power2.size);

	/* set One to the value of 1 */
	One = (BYTE *)calloc(XKey.size, 1);
	One[XKey.size-1] = 0x01;
	
	bcopy((BYTE *)SHA1_IH, tx, SHA1_BYTEHASHLEN);
	w = (BYTE *)calloc(SHA1_BYTEHASHLEN, f);
	XVal = (BYTE *)calloc(XKey.size, 1);

	for ( i=0; i<f; i++ ) {
	
		/* Compute xval = (Xkey + Xseed) mod 2^keylen */
		/* xval = Xkey */
		bzero(XVal, sizeof(XVal));
		bcopy(XKey.val, XVal, XKey.size);

		/* xval = xval + Xseed */
		add(XVal, XKey.size, XSeed.val, XSeed.size);

		/* xval = xval mod 2^keylen */
		Mod(XVal, XKey.size, Power2.val, Power2.size);

		/* Compute w[i] = G(t, xval) */

		offset = b % 8;
		for ( j=0; j<offset; j++ )
			bshl(XVal, XKey.size);

		switch(rndtype) {
		case RAN_DES:
			desG(tx, XVal, &w[i*SHA1_BYTEHASHLEN]);
			break;
		case RAN_SHS:
		default:
			//shaG(tx, XVal, XKey.size, &GfromSHA, SHA1_BYTEHASHLEN);
			shaG(tx, XVal, b, &G, 160);
			bcopy(G, &w[i*SHA1_BYTEHASHLEN], 20);
			FREE(G);
			break;
		}
		/***********************************************
		 * Update XKEY:                                *
		 *   Compute XKEY = (1+XKEY+w[i]) mod 2^keylen *
		 ***********************************************/

		/* XKey = XKey + w_i */
		add(XKey.val, XKey.size, &w[i*SHA1_BYTEHASHLEN], SHA1_BYTEHASHLEN);
		/* XKey = XKey + 1 */
		add(XKey.val, XKey.size, One, XKey.size);

		/* XKey = XKey mod 2^keylen */
		Mod(XKey.val, XKey.size, Power2.val, Power2.size);
	}

	/* set One to the value of 1 */
	if ( One )
		free(One);
	One = (BYTE *)calloc(N.size, 1);
	One[N.size-1] = 0x01;

	Temp = (BYTE *)calloc(N.size, 1);
	bcopy(N.val, Temp, N.size);
	sub(Temp, N.size, One, N.size);

	Mod(w, f*SHA1_BYTEHASHLEN, Temp, N.size);

	allocMP(X, N.bitlen);

	//  w is f*SHA1_BYTEHASHLEN bytes,  I want N.size bytes
	bcopy(w+f*SHA1_BYTEHASHLEN-N.size, X->val, X->size);
	add(X->val, X->size, One, N.size);

	free(XVal);
	free(One);
	free(Temp);
	free(w);
	freeMP(&Power2);
}

//#define GEN_REF_DATA

int ECDSARngTest(int rngtype)
{
	MP	N, XKey, XSeed, X;
#ifdef GEN_REF_DATA
FILE	*fp;
fp = fopen("atest.txt", "a");
#endif

	/*  P-192 Curve test  */
	initMP(&N);
	initMP(&XKey);
	initMP(&XSeed);
	initMP(&X);

	allocMP(&XKey, 160);
	allocMP(&XSeed, 160);

	allocMP(&N, 192);
	// ffff ffff ffff ffff ffff ffff 99de f836 146b c9b1 b4d2 2831
	N.val[0] = 0xff; N.val[1] = 0xff;
	N.val[2] = 0xff; N.val[3] = 0xff;
	N.val[4] = 0xff; N.val[5] = 0xff;
	N.val[6] = 0xff; N.val[7] = 0xff;
	N.val[8] = 0xff; N.val[9] = 0xff;
	N.val[10] = 0xff; N.val[11] = 0xff;
	N.val[12] = 0x99; N.val[13] = 0xde;
	N.val[14] = 0xf8; N.val[15] = 0x36;
	N.val[16] = 0x14; N.val[17] = 0x6b;
	N.val[18] = 0xc9; N.val[19] = 0xb1;
	N.val[20] = 0xb4; N.val[21] = 0xd2;
	N.val[22] = 0x28; N.val[23] = 0x31;
#ifdef GEN_REF_DATA
fprintBstr(fp, "\n[P-192]\nN is ", N.val, N.size);
#endif
		
	/*  Test ECDSA X9.62 RNG - SHA-Gfnc  */
	XKey.val[0] = 0xbd; XKey.val[1] = 0x02;
	XKey.val[2] = 0x9b; XKey.val[3] = 0xbe;
	XKey.val[4] = 0x7f; XKey.val[5] = 0x51;
	XKey.val[6] = 0x96; XKey.val[7] = 0x0b;
	XKey.val[8] = 0xcf; XKey.val[9] = 0x9e;
	XKey.val[10] = 0xdb; XKey.val[11] = 0x2b;
	XKey.val[12] = 0x61; XKey.val[13] = 0xf0;
	XKey.val[14] = 0x6f; XKey.val[15] = 0x0f;
	XKey.val[16] = 0xeb; XKey.val[17] = 0x5a;
	XKey.val[18] = 0x38; XKey.val[19] = 0xb6;
#ifdef GEN_REF_DATA
fprintBstr(fp, "XKey is ", XKey.val, XKey.size);
#endif
	bzero(XSeed.val, XSeed.size);

	ECDSAGenX(XKey, XSeed, 160, N, &X, rngtype);
	
#ifdef GEN_REF_DATA
	if ( rngtype == RAN_SHS )
fprintBstr(fp, "X-SHA is ", X.val, X.size);
	else if ( rngtype == RAN_DES )
fprintBstr(fp, "X-DES is ", X.val, X.size);
	else
		return 20;
fprintBstr(fp, "Updated XKey is ", XKey.val, XKey.size);
#endif


	switch(rngtype) {
	case RAN_SHS:
		// f264 ff1e aaea cdc8 1760 b202 823f b042 5bf0 56b4 e059 08d7
		if ( (X.val[0] != 0xf2) || (X.val[1] != 0x64) ||
			 (X.val[2] != 0xff) || (X.val[3] != 0x1e) ||
			 (X.val[4] != 0xaa) || (X.val[5] != 0xea) ||
			 (X.val[6] != 0xcd) || (X.val[7] != 0xc8) ||
			 (X.val[8] != 0x17) || (X.val[9] != 0x60) ||
			 (X.val[10] != 0xb2) || (X.val[11] != 0x02) ||
			 (X.val[12] != 0x82) || (X.val[13] != 0x3f) ||
			 (X.val[14] != 0xb0) || (X.val[15] != 0x42) ||
			 (X.val[16] != 0x5b) || (X.val[17] != 0xf0) ||
			 (X.val[18] != 0x56) || (X.val[19] != 0xb4) ||
			 (X.val[20] != 0xe0) || (X.val[21] != 0x59) ||
			 (X.val[22] != 0x08) || (X.val[23] != 0xd7) )
			return 1;
		break;
	case RAN_DES:
		// 5069 d527 0069 fe8e 1577 ec3c d49e f3a6 b554 3ee5 00a1 15c9
		if ( (X.val[0] != 0x50) || (X.val[1] != 0x69) ||
			 (X.val[2] != 0xd5) || (X.val[3] != 0x27) ||
			 (X.val[4] != 0x00) || (X.val[5] != 0x69) ||
			 (X.val[6] != 0xfe) || (X.val[7] != 0x8e) ||
			 (X.val[8] != 0x15) || (X.val[9] != 0x77) ||
			 (X.val[10] != 0xec) || (X.val[11] != 0x3c) ||
			 (X.val[12] != 0xd4) || (X.val[13] != 0x9e) ||
			 (X.val[14] != 0xf3) || (X.val[15] != 0xa6) ||
			 (X.val[16] != 0xb5) || (X.val[17] != 0x54) ||
			 (X.val[18] != 0x3e) || (X.val[19] != 0xe5) ||
			 (X.val[20] != 0x00) || (X.val[21] != 0xa1) ||
			 (X.val[22] != 0x15) || (X.val[23] != 0xc9) )
			return 1;
		break;
	default:
		return 20;
	}

	freeMP(&N);
	freeMP(&X);

	/*  P-224 Curve test  */

	allocMP(&N, 224);
	
	// ffff ffff ffff ffff ffff ffff ffff 16a2 e0b8 f03e 13dd 2945 5c5c 2a3d
	N.val[0] = 0xff; N.val[1] = 0xff;
	N.val[2] = 0xff; N.val[3] = 0xff;
	N.val[4] = 0xff; N.val[5] = 0xff;
	N.val[6] = 0xff; N.val[7] = 0xff;
	N.val[8] = 0xff; N.val[9] = 0xff;
	N.val[10] = 0xff; N.val[11] = 0xff;
	N.val[12] = 0xff; N.val[13] = 0xff;
	N.val[14] = 0x16; N.val[15] = 0xa2;
	N.val[16] = 0xe0; N.val[17] = 0xb8;
	N.val[18] = 0xf0; N.val[19] = 0x3e;
	N.val[20] = 0x13; N.val[21] = 0xdd;
	N.val[22] = 0x29; N.val[23] = 0x45;
	N.val[24] = 0x5c; N.val[25] = 0x5c;
	N.val[26] = 0x2a; N.val[27] = 0x3d;
#ifdef GEN_REF_DATA
fprintBstr(fp, "\n[P-224]\nN is ", N.val, N.size);
#endif
		
	/*  Test ECDSA X9.62 RNG - SHA-Gfnc  */
	XKey.val[0] = 0xbd; XKey.val[1] = 0x02;
	XKey.val[2] = 0x9b; XKey.val[3] = 0xbe;
	XKey.val[4] = 0x7f; XKey.val[5] = 0x51;
	XKey.val[6] = 0x96; XKey.val[7] = 0x0b;
	XKey.val[8] = 0xcf; XKey.val[9] = 0x9e;
	XKey.val[10] = 0xdb; XKey.val[11] = 0x2b;
	XKey.val[12] = 0x61; XKey.val[13] = 0xf0;
	XKey.val[14] = 0x6f; XKey.val[15] = 0x0f;
	XKey.val[16] = 0xeb; XKey.val[17] = 0x5a;
	XKey.val[18] = 0x38; XKey.val[19] = 0xb6;
#ifdef GEN_REF_DATA
fprintBstr(fp, "XKey is ", XKey.val, XKey.size);
#endif
	bzero(XSeed.val, XSeed.size);

	ECDSAGenX(XKey, XSeed, 160, N, &X, rngtype);
	
#ifdef GEN_REF_DATA
	if ( rngtype == RAN_SHS )
fprintBstr(fp, "X-SHA is ", X.val, X.size);
	else if ( rngtype == RAN_DES )
fprintBstr(fp, "X-DES is ", X.val, X.size);
	else
		return 20;
fprintBstr(fp, "Updated XKey is ", XKey.val, XKey.size);
#endif

	switch(rngtype) {
	case RAN_SHS:
		// 7b2e 58db eb17 eaa5 26a7 bf59 ca6a 28b1 35b1 e1a5 a3bc 7ace ab64 3a07
		if ( (X.val[0] != 0x7b) || (X.val[1] != 0x2e) ||
			 (X.val[2] != 0x58) || (X.val[3] != 0xdb) ||
			 (X.val[4] != 0xeb) || (X.val[5] != 0x17) ||
			 (X.val[6] != 0xea) || (X.val[7] != 0xa5) ||
			 (X.val[8] != 0x26) || (X.val[9] != 0xa7) ||
			 (X.val[10] != 0xbf) || (X.val[11] != 0x59) ||
			 (X.val[12] != 0xca) || (X.val[13] != 0x6a) ||
			 (X.val[14] != 0x28) || (X.val[15] != 0xb1) ||
			 (X.val[16] != 0x35) || (X.val[17] != 0xb1) ||
			 (X.val[18] != 0xe1) || (X.val[19] != 0xa5) ||
			 (X.val[20] != 0xa3) || (X.val[21] != 0xbc) ||
			 (X.val[22] != 0x7a) || (X.val[23] != 0xce) ||
			 (X.val[24] != 0xab) || (X.val[25] != 0x64) ||
			 (X.val[26] != 0x3a) || (X.val[27] != 0x07) )
			return 2;
		break;
	case RAN_DES:
		// fe04 0c5e 71b3 c7fb 3da5 94ca 62f0
		if ( (X.val[0] != 0xfe) || (X.val[1] != 0x04) ||
			 (X.val[2] != 0x0c) || (X.val[3] != 0x5e) ||
			 (X.val[4] != 0x71) || (X.val[5] != 0xb3) ||
			 (X.val[6] != 0xc7) || (X.val[7] != 0xfb) ||
			 (X.val[8] != 0x3d) || (X.val[9] != 0xa5) ||
			 (X.val[10] != 0x94) || (X.val[11] != 0xca) ||
			 (X.val[12] != 0x62) || (X.val[13] != 0xf0) ||
			 // 13a3 57a5 b730 7e14 0ffd ce9d 8aed
			 (X.val[14] != 0x13) || (X.val[15] != 0xa3) ||
			 (X.val[16] != 0x57) || (X.val[17] != 0xa5) ||
			 (X.val[18] != 0xb7) || (X.val[19] != 0x30) ||
			 (X.val[20] != 0x7e) || (X.val[21] != 0x14) ||
			 (X.val[22] != 0x0f) || (X.val[23] != 0xfd) ||
			 (X.val[24] != 0xce) || (X.val[25] != 0x9d) ||
			 (X.val[26] != 0x8a) || (X.val[27] != 0xed) )
			return 2;
		break;
	default:
		return 20;
	}

	freeMP(&N);
	freeMP(&X);

	/*  P-256 Curve test  */

	allocMP(&N, 256);
	
	// ffff ffff 0000 0000 ffff ffff ffff ffff bce6 faad a717 9e84 f3b9 cac2 fc63 2551
	N.val[0] = 0xff; N.val[1] = 0xff;
	N.val[2] = 0xff; N.val[3] = 0xff;
	N.val[4] = 0x00; N.val[5] = 0x00;
	N.val[6] = 0x00; N.val[7] = 0x00;
	N.val[8] = 0xff; N.val[9] = 0xff;
	N.val[10] = 0xff; N.val[11] = 0xff;
	N.val[12] = 0xff; N.val[13] = 0xff;
	N.val[14] = 0xff; N.val[15] = 0xff;
	N.val[16] = 0xbc; N.val[17] = 0xe6;
	N.val[18] = 0xfa; N.val[19] = 0xad;
	N.val[20] = 0xa7; N.val[21] = 0x17;
	N.val[22] = 0x9e; N.val[23] = 0x84;
	N.val[24] = 0xf3; N.val[25] = 0xb9;
	N.val[26] = 0xca; N.val[27] = 0xc2;
	N.val[28] = 0xfc; N.val[29] = 0x63;
	N.val[30] = 0x25; N.val[31] = 0x51;
#ifdef GEN_REF_DATA
fprintBstr(fp, "\n[P-256]\nN is ", N.val, N.size);
#endif
		
	/*  Test ECDSA X9.62 RNG - SHA-Gfnc  */
	XKey.val[0] = 0xbd; XKey.val[1] = 0x02;
	XKey.val[2] = 0x9b; XKey.val[3] = 0xbe;
	XKey.val[4] = 0x7f; XKey.val[5] = 0x51;
	XKey.val[6] = 0x96; XKey.val[7] = 0x0b;
	XKey.val[8] = 0xcf; XKey.val[9] = 0x9e;
	XKey.val[10] = 0xdb; XKey.val[11] = 0x2b;
	XKey.val[12] = 0x61; XKey.val[13] = 0xf0;
	XKey.val[14] = 0x6f; XKey.val[15] = 0x0f;
	XKey.val[16] = 0xeb; XKey.val[17] = 0x5a;
	XKey.val[18] = 0x38; XKey.val[19] = 0xb6;
#ifdef GEN_REF_DATA
fprintBstr(fp, "XKey is ", XKey.val, XKey.size);
#endif
	bzero(XSeed.val, XSeed.size);

	ECDSAGenX(XKey, XSeed, 160, N, &X, rngtype);
	
#ifdef GEN_REF_DATA
	if ( rngtype == RAN_SHS )
fprintBstr(fp, "X-SHA is ", X.val, X.size);
	else if ( rngtype == RAN_DES )
fprintBstr(fp, "X-DES is ", X.val, X.size);
	else
		return 20;
fprintBstr(fp, "Updated XKey is ", XKey.val, XKey.size);
#endif

	switch(rngtype) {
	case RAN_SHS:
		// 1bd6 472c 1d03 50f7 93a6 b0a4 5c49 c328
		if ( (X.val[0] != 0x1b) || (X.val[1] != 0xd6) ||
			 (X.val[2] != 0x47) || (X.val[3] != 0x2c) ||
			 (X.val[4] != 0x1d) || (X.val[5] != 0x03) ||
			 (X.val[6] != 0x50) || (X.val[7] != 0xf7) ||
			 (X.val[8] != 0x93) || (X.val[9] != 0xa6) ||
			 (X.val[10] != 0xb0) || (X.val[11] != 0xa4) ||
			 (X.val[12] != 0x5c) || (X.val[13] != 0x49) ||
			 (X.val[14] != 0xc3) || (X.val[15] != 0x28) ||

			 // 967c 7131 6ea8 7cc2 a15b f578 ac2a 3d77
			 (X.val[16] != 0x96) || (X.val[17] != 0x7c) ||
			 (X.val[18] != 0x71) || (X.val[19] != 0x31) ||
			 (X.val[20] != 0x6e) || (X.val[21] != 0xa8) ||
			 (X.val[22] != 0x7c) || (X.val[23] != 0xc2) ||
			 (X.val[24] != 0xa1) || (X.val[25] != 0x5b) ||
			 (X.val[26] != 0xf5) || (X.val[27] != 0x78) ||
			 (X.val[28] != 0xac) || (X.val[29] != 0x2a) ||
			 (X.val[30] != 0x3d) || (X.val[31] != 0x77) )
			return 3;
		break;
	case RAN_DES:
		// e1e8 0ec8 1c63 e21e 4ebf dce4 0b5e 3ec5
		if ( (X.val[0] != 0xe1) || (X.val[1] != 0xe8) ||
			 (X.val[2] != 0x0e) || (X.val[3] != 0xc8) ||
			 (X.val[4] != 0x1c) || (X.val[5] != 0x63) ||
			 (X.val[6] != 0xe2) || (X.val[7] != 0x1e) ||
			 (X.val[8] != 0x4e) || (X.val[9] != 0xbf) ||
			 (X.val[10] != 0xdc) || (X.val[11] != 0xe4) ||
			 (X.val[12] != 0x0b) || (X.val[13] != 0x5e) ||
			 (X.val[14] != 0x3e) || (X.val[15] != 0xc5) ||

			 // 391c 3523 d30d 61f6 0720 7ad7 eaf3 b7a9
			 (X.val[16] != 0x39) || (X.val[17] != 0x1c) ||
			 (X.val[18] != 0x35) || (X.val[19] != 0x23) ||
			 (X.val[20] != 0xd3) || (X.val[21] != 0x0d) ||
			 (X.val[22] != 0x61) || (X.val[23] != 0xf6) ||
			 (X.val[24] != 0x07) || (X.val[25] != 0x20) ||
			 (X.val[26] != 0x7a) || (X.val[27] != 0xd7) ||
			 (X.val[28] != 0xea) || (X.val[29] != 0xf3) ||
			 (X.val[30] != 0xb7) || (X.val[31] != 0xa9) )
			return 3;
		break;
	default:
		return 20;
	}

	freeMP(&N);
	freeMP(&X);

	/*  P-384 Curve test  */

	allocMP(&N, 384);
	
	// ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff
	N.val[0] = 0xff; N.val[1] = 0xff;
	N.val[2] = 0xff; N.val[3] = 0xff;
	N.val[4] = 0xff; N.val[5] = 0xff;
	N.val[6] = 0xff; N.val[7] = 0xff;
	N.val[8] = 0xff; N.val[9] = 0xff;
	N.val[10] = 0xff; N.val[11] = 0xff;
	N.val[12] = 0xff; N.val[13] = 0xff;
	N.val[14] = 0xff; N.val[15] = 0xff;
	N.val[16] = 0xff; N.val[17] = 0xff;
	N.val[18] = 0xff; N.val[19] = 0xff;
	N.val[20] = 0xff; N.val[21] = 0xff;
	N.val[22] = 0xff; N.val[23] = 0xff;

	// c763 4d81 f437 2ddf 581a 0db2 48b0 a77a ecec 196a ccc5 2973
	N.val[24] = 0xc7; N.val[25] = 0x63;
	N.val[26] = 0x4d; N.val[27] = 0x81;
	N.val[28] = 0xf4; N.val[29] = 0x37;
	N.val[30] = 0x2d; N.val[31] = 0xdf;
	N.val[32] = 0x58; N.val[33] = 0x1a;
	N.val[34] = 0x0d; N.val[35] = 0xb2;
	N.val[36] = 0x48; N.val[37] = 0xb0;
	N.val[38] = 0xa7; N.val[39] = 0x7a;
	N.val[40] = 0xec; N.val[41] = 0xec;
	N.val[42] = 0x19; N.val[43] = 0x6a;
	N.val[44] = 0xcc; N.val[45] = 0xc5;
	N.val[46] = 0x29; N.val[47] = 0x73;
#ifdef GEN_REF_DATA
fprintBstr(fp, "\n[P-384]\nN is ", N.val, N.size);
#endif
		
	/*  Test ECDSA X9.62 RNG - SHA-Gfnc  */
	XKey.val[0] = 0xbd; XKey.val[1] = 0x02;
	XKey.val[2] = 0x9b; XKey.val[3] = 0xbe;
	XKey.val[4] = 0x7f; XKey.val[5] = 0x51;
	XKey.val[6] = 0x96; XKey.val[7] = 0x0b;
	XKey.val[8] = 0xcf; XKey.val[9] = 0x9e;
	XKey.val[10] = 0xdb; XKey.val[11] = 0x2b;
	XKey.val[12] = 0x61; XKey.val[13] = 0xf0;
	XKey.val[14] = 0x6f; XKey.val[15] = 0x0f;
	XKey.val[16] = 0xeb; XKey.val[17] = 0x5a;
	XKey.val[18] = 0x38; XKey.val[19] = 0xb6;
#ifdef GEN_REF_DATA
fprintBstr(fp, "XKey is ", XKey.val, XKey.size);
#endif
	bzero(XSeed.val, XSeed.size);

	ECDSAGenX(XKey, XSeed, 160, N, &X, rngtype);
	
#ifdef GEN_REF_DATA
	if ( rngtype == RAN_SHS )
fprintBstr(fp, "X-SHA is ", X.val, X.size);
	else if ( rngtype == RAN_DES )
fprintBstr(fp, "X-DES is ", X.val, X.size);
	else
		return 20;
fprintBstr(fp, "Updated XKey is ", XKey.val, XKey.size);
#endif

	switch(rngtype) {
	case RAN_SHS:
		// 7b2e 3b49 8b26 0614 3c6c 18ba d23b eed0 
		if ( (X.val[0] != 0x7b) || (X.val[1] != 0x2e) ||
			 (X.val[2] != 0x3b) || (X.val[3] != 0x49) ||
			 (X.val[4] != 0x8b) || (X.val[5] != 0x26) ||
			 (X.val[6] != 0x06) || (X.val[7] != 0x14) ||
			 (X.val[8] != 0x3c) || (X.val[9] != 0x6c) ||
			 (X.val[10] != 0x18) || (X.val[11] != 0xba) ||
			 (X.val[12] != 0xd2) || (X.val[13] != 0x3b) ||
			 (X.val[14] != 0xee) || (X.val[15] != 0xd0) ||

			 // b69d b129 859f 22f3 f07d db30 708b cd60
			 (X.val[16] != 0xb6) || (X.val[17] != 0x9d) ||
			 (X.val[18] != 0xb1) || (X.val[19] != 0x29) ||
			 (X.val[20] != 0x85) || (X.val[21] != 0x9f) ||
			 (X.val[22] != 0x22) || (X.val[23] != 0xf3) ||
			 (X.val[24] != 0xf0) || (X.val[25] != 0x7d) ||
			 (X.val[26] != 0xdb) || (X.val[27] != 0x30) ||
			 (X.val[28] != 0x70) || (X.val[29] != 0x8b) ||
			 (X.val[30] != 0xcd) || (X.val[31] != 0x60) ||

			 // 8bc3 e274 3260 0030 e373 afeb 9ce1 697f
			 (X.val[32] != 0x8b) || (X.val[33] != 0xc3) ||
			 (X.val[34] != 0xe2) || (X.val[35] != 0x74) ||
			 (X.val[36] != 0x32) || (X.val[37] != 0x60) ||
			 (X.val[38] != 0x00) || (X.val[39] != 0x30) ||
			 (X.val[40] != 0xe3) || (X.val[41] != 0x73) ||
			 (X.val[42] != 0xaf) || (X.val[43] != 0xeb) ||
			 (X.val[44] != 0x9c) || (X.val[45] != 0xe1) ||
			 (X.val[46] != 0x69) || (X.val[47] != 0x7f) )
			return 4;
		break;
	case RAN_DES:
		// fe03 36ff 1166 516e a319 8afc d7c3 03c5
		if ( (X.val[0] != 0xfe) || (X.val[1] != 0x03) ||
			 (X.val[2] != 0x36) || (X.val[3] != 0xff) ||
			 (X.val[4] != 0x11) || (X.val[5] != 0x66) ||
			 (X.val[6] != 0x51) || (X.val[7] != 0x6e) ||
			 (X.val[8] != 0xa3) || (X.val[9] != 0x19) ||
			 (X.val[10] != 0x8a) || (X.val[11] != 0xfc) ||
			 (X.val[12] != 0xd7) || (X.val[13] != 0xc3) ||
			 (X.val[14] != 0x03) || (X.val[15] != 0xc5) ||

			 // 1f35 876a c5e6 334c 81b3 04ea b28c 9883
			 (X.val[16] != 0x1f) || (X.val[17] != 0x35) ||
			 (X.val[18] != 0x87) || (X.val[19] != 0x6a) ||
			 (X.val[20] != 0xc5) || (X.val[21] != 0xe6) ||
			 (X.val[22] != 0x33) || (X.val[23] != 0x4c) ||
			 (X.val[24] != 0x81) || (X.val[25] != 0xb3) ||
			 (X.val[26] != 0x04) || (X.val[27] != 0xea) ||
			 (X.val[28] != 0xb2) || (X.val[29] != 0x8c) ||
			 (X.val[30] != 0x98) || (X.val[31] != 0x83) ||

			 // 01a8 a9f8 8025 3c7f 8f7c b3c7 77c2 f860
			 (X.val[32] != 0x01) || (X.val[33] != 0xa8) ||
			 (X.val[34] != 0xa9) || (X.val[35] != 0xf8) ||
			 (X.val[36] != 0x80) || (X.val[37] != 0x25) ||
			 (X.val[38] != 0x3c) || (X.val[39] != 0x7f) ||
			 (X.val[40] != 0x8f) || (X.val[41] != 0x7c) ||
			 (X.val[42] != 0xb3) || (X.val[43] != 0xc7) ||
			 (X.val[44] != 0x77) || (X.val[45] != 0xc2) ||
			 (X.val[46] != 0xf8) || (X.val[47] != 0x60) )
			return 4;
		break;
	default:
		return 20;
	}

	freeMP(&N);
	freeMP(&X);

	/*  P-521 Curve test  */

	allocMP(&N, 521);
	
	// 01ff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff
	// fffa 5186 8783 bf2f 966b 7fcc 0148 f709 a5d0 3bb5 c9b8 899c 47ae bb6f b71e 9138 
	// 6409
	N.val[0] = 0x01; N.val[1] = 0xff;
	N.val[2] = 0xff; N.val[3] = 0xff;
	N.val[4] = 0xff; N.val[5] = 0xff;
	N.val[6] = 0xff; N.val[7] = 0xff;
	N.val[8] = 0xff; N.val[9] = 0xff;
	N.val[10] = 0xff; N.val[11] = 0xff;
	N.val[12] = 0xff; N.val[13] = 0xff;
	N.val[14] = 0xff; N.val[15] = 0xff;
	N.val[16] = 0xff; N.val[17] = 0xff;
	N.val[18] = 0xff; N.val[19] = 0xff;
	N.val[20] = 0xff; N.val[21] = 0xff;
	N.val[22] = 0xff; N.val[23] = 0xff;
	N.val[24] = 0xff; N.val[25] = 0xff;
	N.val[26] = 0xff; N.val[27] = 0xff;
	N.val[28] = 0xff; N.val[29] = 0xff;
	N.val[30] = 0xff; N.val[31] = 0xff;

	N.val[32] = 0xff; N.val[33] = 0xfa;
	N.val[34] = 0x51; N.val[35] = 0x86;
	N.val[36] = 0x87; N.val[37] = 0x83;
	N.val[38] = 0xbf; N.val[39] = 0x2f;
	N.val[40] = 0x96; N.val[41] = 0x6b;
	N.val[42] = 0x7f; N.val[43] = 0xcc;
	N.val[44] = 0x01; N.val[45] = 0x48;
	N.val[46] = 0xf7; N.val[47] = 0x09;
	N.val[48] = 0xa5; N.val[49] = 0xd0;
	N.val[50] = 0x3b; N.val[51] = 0xb5;
	N.val[52] = 0xc9; N.val[53] = 0xb8;
	N.val[54] = 0x89; N.val[55] = 0x9c;
	N.val[56] = 0x47; N.val[57] = 0xae;
	N.val[58] = 0xbb; N.val[59] = 0x6f;
	N.val[60] = 0xb7; N.val[61] = 0x1e;
	N.val[62] = 0x91; N.val[63] = 0x38;

	N.val[64] = 0x64; N.val[65] = 0x09;
#ifdef GEN_REF_DATA
fprintBstr(fp, "\n[P-521]\nN is ", N.val, N.size);
#endif
		
	/*  Test ECDSA X9.62 RNG - SHA-Gfnc  */
	XKey.val[0] = 0xbd; XKey.val[1] = 0x02;
	XKey.val[2] = 0x9b; XKey.val[3] = 0xbe;
	XKey.val[4] = 0x7f; XKey.val[5] = 0x51;
	XKey.val[6] = 0x96; XKey.val[7] = 0x0b;
	XKey.val[8] = 0xcf; XKey.val[9] = 0x9e;
	XKey.val[10] = 0xdb; XKey.val[11] = 0x2b;
	XKey.val[12] = 0x61; XKey.val[13] = 0xf0;
	XKey.val[14] = 0x6f; XKey.val[15] = 0x0f;
	XKey.val[16] = 0xeb; XKey.val[17] = 0x5a;
	XKey.val[18] = 0x38; XKey.val[19] = 0xb6;
#ifdef GEN_REF_DATA
fprintBstr(fp, "XKey is ", XKey.val, XKey.size);
#endif
	bzero(XSeed.val, XSeed.size);

	ECDSAGenX(XKey, XSeed, 160, N, &X, rngtype);
	
#ifdef GEN_REF_DATA
	if ( rngtype == RAN_SHS )
fprintBstr(fp, "X-SHA is ", X.val, X.size);
	else if ( rngtype == RAN_DES )
fprintBstr(fp, "X-DES is ", X.val, X.size);
	else
		return 20;
fprintBstr(fp, "Updated XKey is ", XKey.val, XKey.size);
#endif

	switch(rngtype) {
	case RAN_SHS:
		// 0149 8b26 0614 3c6c 18ba cb0f 6c55 babb
		if ( (X.val[0] != 0x01) || (X.val[1] != 0x49) ||
			 (X.val[2] != 0x8b) || (X.val[3] != 0x26) ||
			 (X.val[4] != 0x06) || (X.val[5] != 0x14) ||
			 (X.val[6] != 0x3c) || (X.val[7] != 0x6c) ||
			 (X.val[8] != 0x18) || (X.val[9] != 0xba) ||
			 (X.val[10] != 0xcb) || (X.val[11] != 0x0f) ||
			 (X.val[12] != 0x6c) || (X.val[13] != 0x55) ||
			 (X.val[14] != 0xba) || (X.val[15] != 0xbb) ||

			 // 1378 8e7c fef6 9e03 9733 a512 0694 2161
			 (X.val[16] != 0x13) || (X.val[17] != 0x78) ||
			 (X.val[18] != 0x8e) || (X.val[19] != 0x7c) ||
			 (X.val[20] != 0xfe) || (X.val[21] != 0xf6) ||
			 (X.val[22] != 0x9e) || (X.val[23] != 0x03) ||
			 (X.val[24] != 0x97) || (X.val[25] != 0x33) ||
			 (X.val[26] != 0xa5) || (X.val[27] != 0x12) ||
			 (X.val[28] != 0x06) || (X.val[29] != 0x94) ||
			 (X.val[30] != 0x21) || (X.val[31] != 0x61) ||

			 // bc2a e6f4 5f90 8f48 77a1 ceb9 4a5a 2b69
			 (X.val[32] != 0xbc) || (X.val[33] != 0x2a) ||
			 (X.val[34] != 0xe6) || (X.val[35] != 0xf4) ||
			 (X.val[36] != 0x5f) || (X.val[37] != 0x90) ||
			 (X.val[38] != 0x8f) || (X.val[39] != 0x48) ||
			 (X.val[40] != 0x77) || (X.val[41] != 0xa1) ||
			 (X.val[42] != 0xce) || (X.val[43] != 0xb9) ||
			 (X.val[44] != 0x4a) || (X.val[45] != 0x5a) ||
			 (X.val[46] != 0x2b) || (X.val[47] != 0x69) ||

			 // 40d3 b58a 15eb c220 1a07 3c72 0f31 ebee dd74
			 (X.val[48] != 0x40) || (X.val[49] != 0xd3) ||
			 (X.val[50] != 0xb5) || (X.val[51] != 0x8a) ||
			 (X.val[52] != 0x15) || (X.val[53] != 0xeb) ||
			 (X.val[54] != 0xc2) || (X.val[55] != 0x20) ||
			 (X.val[56] != 0x1a) || (X.val[57] != 0x07) ||
			 (X.val[58] != 0x3c) || (X.val[59] != 0x72) ||
			 (X.val[60] != 0x0f) || (X.val[61] != 0x31) ||
			 (X.val[62] != 0xeb) || (X.val[63] != 0xee) ||
			 (X.val[64] != 0xdd) || (X.val[65] != 0x74) )
			return 5;
		break;
	case RAN_DES:
		// 00ff 1166 516e a319 8afc a3ff d705 20fd
		if ( (X.val[0] != 0x00) || (X.val[1] != 0xff) ||
			 (X.val[2] != 0x11) || (X.val[3] != 0x66) ||
			 (X.val[4] != 0x51) || (X.val[5] != 0x6e) ||
			 (X.val[6] != 0xa3) || (X.val[7] != 0x19) ||
			 (X.val[8] != 0x8a) || (X.val[9] != 0xfc) ||
			 (X.val[10] != 0xa3) || (X.val[11] != 0xff) ||
			 (X.val[12] != 0xd7) || (X.val[13] != 0x05) ||
			 (X.val[14] != 0x20) || (X.val[15] != 0xfd) ||

			 // 68c3 2d56 ee75 1382 55e5 16dd 3349 828d
			 (X.val[16] != 0x68) || (X.val[17] != 0xc3) ||
			 (X.val[18] != 0x2d) || (X.val[19] != 0x56) ||
			 (X.val[20] != 0xee) || (X.val[21] != 0x75) ||
			 (X.val[22] != 0x13) || (X.val[23] != 0x82) ||
			 (X.val[24] != 0x55) || (X.val[25] != 0xe5) ||
			 (X.val[26] != 0x16) || (X.val[27] != 0xdd) ||
			 (X.val[28] != 0x33) || (X.val[29] != 0x49) ||
			 (X.val[30] != 0x82) || (X.val[31] != 0x8d) ||

			 // b7c2 0c28 b90a ce32 1edf e47f 506c 8dc7
			 (X.val[32] != 0xb7) || (X.val[33] != 0xc2) ||
			 (X.val[34] != 0x0c) || (X.val[35] != 0x28) ||
			 (X.val[36] != 0xb9) || (X.val[37] != 0x0a) ||
			 (X.val[38] != 0xce) || (X.val[39] != 0x32) ||
			 (X.val[40] != 0x1e) || (X.val[41] != 0xdf) ||
			 (X.val[42] != 0xe4) || (X.val[43] != 0x7f) ||
			 (X.val[44] != 0x50) || (X.val[45] != 0x6c) ||
			 (X.val[46] != 0x8d) || (X.val[47] != 0xc7) ||

			 // 5c91 0b24 c50f 13d0 184c 9efc f415 9433 f1d1
			 (X.val[48] != 0x5c) || (X.val[49] != 0x91) ||
			 (X.val[50] != 0x0b) || (X.val[51] != 0x24) ||
			 (X.val[52] != 0xc5) || (X.val[53] != 0x0f) ||
			 (X.val[54] != 0x13) || (X.val[55] != 0xd0) ||
			 (X.val[56] != 0x18) || (X.val[57] != 0x4c) ||
			 (X.val[58] != 0x9e) || (X.val[59] != 0xfc) ||
			 (X.val[60] != 0xf4) || (X.val[61] != 0x15) ||
			 (X.val[62] != 0x94) || (X.val[63] != 0x33) ||
			 (X.val[64] != 0xf1) || (X.val[65] != 0xd1) )
			return 5;
		break;
	default:
		return 20;
	}

	freeMP(&N);
	freeMP(&X);

	/*  K-163 Curve test  */

	allocMP(&N, 163);
	
	// 0400 0000 0000 0000 0000 0201 08a2 e0cc 0d99 f8a5 ef
	N.val[0] = 0x04; N.val[1] = 0x00;
	N.val[2] = 0x00; N.val[3] = 0x00;
	N.val[4] = 0x00; N.val[5] = 0x00;
	N.val[6] = 0x00; N.val[7] = 0x00;
	N.val[8] = 0x00; N.val[9] = 0x00;
	N.val[10] = 0x02; N.val[11] = 0x01;
	N.val[12] = 0x08; N.val[13] = 0xa2;
	N.val[14] = 0xe0; N.val[15] = 0xcc;
	N.val[16] = 0x0d; N.val[17] = 0x99;
	N.val[18] = 0xf8; N.val[19] = 0xa5;
	N.val[20] = 0xef;
#ifdef GEN_REF_DATA
fprintBstr(fp, "\n[K-163]\nN is ", N.val, N.size);
#endif
		
	/*  Test ECDSA X9.62 RNG - SHA-Gfnc  */
	XKey.val[0] = 0xbd; XKey.val[1] = 0x02;
	XKey.val[2] = 0x9b; XKey.val[3] = 0xbe;
	XKey.val[4] = 0x7f; XKey.val[5] = 0x51;
	XKey.val[6] = 0x96; XKey.val[7] = 0x0b;
	XKey.val[8] = 0xcf; XKey.val[9] = 0x9e;
	XKey.val[10] = 0xdb; XKey.val[11] = 0x2b;
	XKey.val[12] = 0x61; XKey.val[13] = 0xf0;
	XKey.val[14] = 0x6f; XKey.val[15] = 0x0f;
	XKey.val[16] = 0xeb; XKey.val[17] = 0x5a;
	XKey.val[18] = 0x38; XKey.val[19] = 0xb6;
#ifdef GEN_REF_DATA
fprintBstr(fp, "XKey is ", XKey.val, XKey.size);
#endif
	bzero(XSeed.val, XSeed.size);

	ECDSAGenX(XKey, XSeed, 160, N, &X, rngtype);
	
#ifdef GEN_REF_DATA
	if ( rngtype == RAN_SHS )
fprintBstr(fp, "X-SHA is ", X.val, X.size);
	else if ( rngtype == RAN_DES )
fprintBstr(fp, "X-DES is ", X.val, X.size);
	else
		return 20;
fprintBstr(fp, "Updated XKey is ", XKey.val, XKey.size);
#endif

	switch(rngtype) {
	case RAN_SHS:
		// 037f 8b9d 304c 587f f9d4 e305 7a17 93cd bc54 e999 eb
		if ( (X.val[0] != 0x03) || (X.val[1] != 0x7f) ||
			 (X.val[2] != 0x8b) || (X.val[3] != 0x9d) ||
			 (X.val[4] != 0x30) || (X.val[5] != 0x4c) ||
			 (X.val[6] != 0x58) || (X.val[7] != 0x7f) ||
			 (X.val[8] != 0xf9) || (X.val[9] != 0xd4) ||
			 (X.val[10] != 0xe3) || (X.val[11] != 0x05) ||
			 (X.val[12] != 0x7a) || (X.val[13] != 0x17) ||
			 (X.val[14] != 0x93) || (X.val[15] != 0xcd) ||
			 (X.val[16] != 0xbc) || (X.val[17] != 0x54) ||
			 (X.val[18] != 0xe9) || (X.val[19] != 0x99) ||
			 (X.val[20] != 0xeb) )
			return 6;
		break;
	case RAN_DES:
		// 01a6 83d7 49a4 a71a df43 19e6 3142 b218 cf01 5a6b f1
		if ( (X.val[0] != 0x01) || (X.val[1] != 0xa6) ||
			 (X.val[2] != 0x83) || (X.val[3] != 0xd7) ||
			 (X.val[4] != 0x49) || (X.val[5] != 0xa4) ||
			 (X.val[6] != 0xa7) || (X.val[7] != 0x1a) ||
			 (X.val[8] != 0xdf) || (X.val[9] != 0x43) ||
			 (X.val[10] != 0x19) || (X.val[11] != 0xe6) ||
			 (X.val[12] != 0x31) || (X.val[13] != 0x42) ||
			 (X.val[14] != 0xb2) || (X.val[15] != 0x18) ||
			 (X.val[16] != 0xcf) || (X.val[17] != 0x01) ||
			 (X.val[18] != 0x5a) || (X.val[19] != 0x6b) ||
			 (X.val[20] != 0xf1) )
			return 6;
		break;
	default:
		return 20;
	}

	freeMP(&X);

	/*  B-163 Curve test  */
	
	// 0400 0000 0000 0000 0000 0292 fe77 e70c 12a4 234c 33
	N.val[0] = 0x04; N.val[1] = 0x00;
	N.val[2] = 0x00; N.val[3] = 0x00;
	N.val[4] = 0x00; N.val[5] = 0x00;
	N.val[6] = 0x00; N.val[7] = 0x00;
	N.val[8] = 0x00; N.val[9] = 0x00;
	N.val[10] = 0x02; N.val[11] = 0x92;
	N.val[12] = 0xfe; N.val[13] = 0x77;
	N.val[14] = 0xe7; N.val[15] = 0x0c;
	N.val[16] = 0x12; N.val[17] = 0xa4;
	N.val[18] = 0x23; N.val[19] = 0x4c;
	N.val[20] = 0x33;
#ifdef GEN_REF_DATA
fprintBstr(fp, "\n[B-163]\nN is ", N.val, N.size);
#endif
		
	/*  Test ECDSA X9.62 RNG - SHA-Gfnc  */
	XKey.val[0] = 0xbd; XKey.val[1] = 0x02;
	XKey.val[2] = 0x9b; XKey.val[3] = 0xbe;
	XKey.val[4] = 0x7f; XKey.val[5] = 0x51;
	XKey.val[6] = 0x96; XKey.val[7] = 0x0b;
	XKey.val[8] = 0xcf; XKey.val[9] = 0x9e;
	XKey.val[10] = 0xdb; XKey.val[11] = 0x2b;
	XKey.val[12] = 0x61; XKey.val[13] = 0xf0;
	XKey.val[14] = 0x6f; XKey.val[15] = 0x0f;
	XKey.val[16] = 0xeb; XKey.val[17] = 0x5a;
	XKey.val[18] = 0x38; XKey.val[19] = 0xb6;
#ifdef GEN_REF_DATA
fprintBstr(fp, "XKey is ", XKey.val, XKey.size);
#endif
	bzero(XSeed.val, XSeed.size);

	ECDSAGenX(XKey, XSeed, 160, N, &X, rngtype);
	
#ifdef GEN_REF_DATA
	if ( rngtype == RAN_SHS )
fprintBstr(fp, "X-SHA is ", X.val, X.size);
	else if ( rngtype == RAN_DES )
fprintBstr(fp, "X-DES is ", X.val, X.size);
	else
		return 20;
fprintBstr(fp, "Updated XKey is ", XKey.val, XKey.size);
#endif

	switch(rngtype) {
	case RAN_SHS:
		// 0013 fd75 2f14 7b3b 728f c77a cc6c 2a0f af81 746f 33
		if ( (X.val[0] != 0x00) || (X.val[1] != 0x13) ||
			 (X.val[2] != 0xfd) || (X.val[3] != 0x75) ||
			 (X.val[4] != 0x2f) || (X.val[5] != 0x14) ||
			 (X.val[6] != 0x7b) || (X.val[7] != 0x3b) ||
			 (X.val[8] != 0x72) || (X.val[9] != 0x8f) ||
			 (X.val[10] != 0xc7) || (X.val[11] != 0x7a) ||
			 (X.val[12] != 0xcc) || (X.val[13] != 0x6c) ||
			 (X.val[14] != 0x2a) || (X.val[15] != 0x0f) ||
			 (X.val[16] != 0xaf) || (X.val[17] != 0x81) ||
			 (X.val[18] != 0x74) || (X.val[19] != 0x6f) ||
			 (X.val[20] != 0x33) )
			return 7;
		break;
	case RAN_DES:
		// 0165 bb9e a378 75de 65a6 46f4 0b5f df97 6322 1b34 9d
		if ( (X.val[0] != 0x01) || (X.val[1] != 0x65) ||
			 (X.val[2] != 0xbb) || (X.val[3] != 0x9e) ||
			 (X.val[4] != 0xa3) || (X.val[5] != 0x78) ||
			 (X.val[6] != 0x75) || (X.val[7] != 0xde) ||
			 (X.val[8] != 0x65) || (X.val[9] != 0xa6) ||
			 (X.val[10] != 0x46) || (X.val[11] != 0xf4) ||
			 (X.val[12] != 0x0b) || (X.val[13] != 0x5f) ||
			 (X.val[14] != 0xdf) || (X.val[15] != 0x97) ||
			 (X.val[16] != 0x63) || (X.val[17] != 0x22) ||
			 (X.val[18] != 0x1b) || (X.val[19] != 0x34) ||
			 (X.val[20] != 0x9d) )
			return 7;
		break;
	default:
		return 20;
	}

	freeMP(&N);
	freeMP(&X);

	/*  K-233 Curve test  */

	allocMP(&N, 233);
	
	// 0080 0000 0000 0000 0000 0000 0000 0006
	N.val[0] = 0x00; N.val[1] = 0x80;
	N.val[2] = 0x00; N.val[3] = 0x00;
	N.val[4] = 0x00; N.val[5] = 0x00;
	N.val[6] = 0x00; N.val[7] = 0x00;
	N.val[8] = 0x00; N.val[9] = 0x00;
	N.val[10] = 0x00; N.val[11] = 0x00;
	N.val[12] = 0x00; N.val[13] = 0x00;
	N.val[14] = 0x00; N.val[15] = 0x06;

	// 9d5b b915 bcd4 6efb 1ad5 f173 abdf
	N.val[16] = 0x9d; N.val[17] = 0x5b;
	N.val[18] = 0xb9; N.val[19] = 0x15;
	N.val[20] = 0xbc; N.val[21] = 0xd4;
	N.val[22] = 0x6e; N.val[23] = 0xfb;
	N.val[24] = 0x1a; N.val[25] = 0xd5;
	N.val[26] = 0xf1; N.val[27] = 0x73;
	N.val[28] = 0xab; N.val[29] = 0xdf;
#ifdef GEN_REF_DATA
fprintBstr(fp, "\n[K-233]\nN is ", N.val, N.size);
#endif
		
	/*  Test ECDSA X9.62 RNG - SHA-Gfnc  */
	XKey.val[0] = 0xbd; XKey.val[1] = 0x02;
	XKey.val[2] = 0x9b; XKey.val[3] = 0xbe;
	XKey.val[4] = 0x7f; XKey.val[5] = 0x51;
	XKey.val[6] = 0x96; XKey.val[7] = 0x0b;
	XKey.val[8] = 0xcf; XKey.val[9] = 0x9e;
	XKey.val[10] = 0xdb; XKey.val[11] = 0x2b;
	XKey.val[12] = 0x61; XKey.val[13] = 0xf0;
	XKey.val[14] = 0x6f; XKey.val[15] = 0x0f;
	XKey.val[16] = 0xeb; XKey.val[17] = 0x5a;
	XKey.val[18] = 0x38; XKey.val[19] = 0xb6;
#ifdef GEN_REF_DATA
fprintBstr(fp, "XKey is ", XKey.val, XKey.size);
#endif
	bzero(XSeed.val, XSeed.size);

	ECDSAGenX(XKey, XSeed, 160, N, &X, rngtype);
	
#ifdef GEN_REF_DATA
	if ( rngtype == RAN_SHS )
fprintBstr(fp, "X-SHA is ", X.val, X.size);
	else if ( rngtype == RAN_DES )
fprintBstr(fp, "X-DES is ", X.val, X.size);
	else
		return 20;
fprintBstr(fp, "Updated XKey is ", XKey.val, XKey.size);
#endif

	switch(rngtype) {
	case RAN_SHS:
		// 007c 7b2e 399c 6145 ab70 334d 0da1 9638
		if ( (X.val[0] != 0x00) || (X.val[1] != 0x7c) ||
			 (X.val[2] != 0x7b) || (X.val[3] != 0x2e) ||
			 (X.val[4] != 0x39) || (X.val[5] != 0x9c) ||
			 (X.val[6] != 0x61) || (X.val[7] != 0x45) ||
			 (X.val[8] != 0xab) || (X.val[9] != 0x70) ||
			 (X.val[10] != 0x33) || (X.val[11] != 0x4d) ||
			 (X.val[12] != 0x0d) || (X.val[13] != 0xa1) ||
			 (X.val[14] != 0x96) || (X.val[15] != 0x38) ||

			 // cdfc 07f9 4016 60e4 1b60 7f84 f135
			 (X.val[16] != 0xcd) || (X.val[17] != 0xfc) ||
			 (X.val[18] != 0x07) || (X.val[19] != 0xf9) ||
			 (X.val[20] != 0x40) || (X.val[21] != 0x16) ||
			 (X.val[22] != 0x60) || (X.val[23] != 0xe4) ||
			 (X.val[24] != 0x1b) || (X.val[25] != 0x60) ||
			 (X.val[26] != 0x7f) || (X.val[27] != 0x84) ||
			 (X.val[28] != 0xf1) || (X.val[29] != 0x35) )
			return 8;
		break;
	case RAN_DES:
		// 004d fe03 2ae6 79ca a785 3686 5689 78d2
		if ( (X.val[0] != 0x00) || (X.val[1] != 0x4d) ||
			 (X.val[2] != 0xfe) || (X.val[3] != 0x03) ||
			 (X.val[4] != 0x2a) || (X.val[5] != 0xe6) ||
			 (X.val[6] != 0x79) || (X.val[7] != 0xca) ||
			 (X.val[8] != 0xa7) || (X.val[9] != 0x85) ||
			 (X.val[10] != 0x36) || (X.val[11] != 0x86) ||
			 (X.val[12] != 0x56) || (X.val[13] != 0x89) ||
			 (X.val[14] != 0x78) || (X.val[15] != 0xd2) ||

			 // 2aeb 0fb5 ec12 2bb9 f49d 291b a8f5
			 (X.val[16] != 0x2a) || (X.val[17] != 0xeb) ||
			 (X.val[18] != 0x0f) || (X.val[19] != 0xb5) ||
			 (X.val[20] != 0xec) || (X.val[21] != 0x12) ||
			 (X.val[22] != 0x2b) || (X.val[23] != 0xb9) ||
			 (X.val[24] != 0xf4) || (X.val[25] != 0x9d) ||
			 (X.val[26] != 0x29) || (X.val[27] != 0x1b) ||
			 (X.val[28] != 0xa8) || (X.val[29] != 0xf5) )
			return 8;
		break;
	default:
		return 20;
	}

	freeMP(&X);

	/*  B-233 Curve test  */
	
	// 0100 0000 0000 0000 0000 0000 0000 0013
	N.val[0] = 0x01; N.val[1] = 0x00;
	N.val[2] = 0x00; N.val[3] = 0x00;
	N.val[4] = 0x00; N.val[5] = 0x00;
	N.val[6] = 0x00; N.val[7] = 0x00;
	N.val[8] = 0x00; N.val[9] = 0x00;
	N.val[10] = 0x00; N.val[11] = 0x00;
	N.val[12] = 0x00; N.val[13] = 0x00;
	N.val[14] = 0x00; N.val[15] = 0x13;

	// e974 e72f 8a69 2203 1d26 03cf e0d7
	N.val[16] = 0xe9; N.val[17] = 0x74;
	N.val[18] = 0xe7; N.val[19] = 0x2f;
	N.val[20] = 0x8a; N.val[21] = 0x69;
	N.val[22] = 0x22; N.val[23] = 0x03;
	N.val[24] = 0x1d; N.val[25] = 0x26;
	N.val[26] = 0x03; N.val[27] = 0xcf;
	N.val[28] = 0xe0; N.val[29] = 0xd7;
#ifdef GEN_REF_DATA
fprintBstr(fp, "\n[B-233]\nN is ", N.val, N.size);
#endif
		
	/*  Test ECDSA X9.62 RNG - SHA-Gfnc  */
	XKey.val[0] = 0xbd; XKey.val[1] = 0x02;
	XKey.val[2] = 0x9b; XKey.val[3] = 0xbe;
	XKey.val[4] = 0x7f; XKey.val[5] = 0x51;
	XKey.val[6] = 0x96; XKey.val[7] = 0x0b;
	XKey.val[8] = 0xcf; XKey.val[9] = 0x9e;
	XKey.val[10] = 0xdb; XKey.val[11] = 0x2b;
	XKey.val[12] = 0x61; XKey.val[13] = 0xf0;
	XKey.val[14] = 0x6f; XKey.val[15] = 0x0f;
	XKey.val[16] = 0xeb; XKey.val[17] = 0x5a;
	XKey.val[18] = 0x38; XKey.val[19] = 0xb6;
#ifdef GEN_REF_DATA
fprintBstr(fp, "XKey is ", XKey.val, XKey.size);
#endif
	bzero(XSeed.val, XSeed.size);

	ECDSAGenX(XKey, XSeed, 160, N, &X, rngtype);
	
#ifdef GEN_REF_DATA
	if ( rngtype == RAN_SHS )
fprintBstr(fp, "X-SHA is ", X.val, X.size);
	else if ( rngtype == RAN_DES )
fprintBstr(fp, "X-DES is ", X.val, X.size);
	else
		return 20;
fprintBstr(fp, "Updated XKey is ", XKey.val, XKey.size);
#endif

	switch(rngtype) {
	case RAN_SHS:
		// 00fc 7b2e 38c3 9877 1471 d902 da03 c182
		if ( (X.val[0] != 0x00) || (X.val[1] != 0xfc) ||
			 (X.val[2] != 0x7b) || (X.val[3] != 0x2e) ||
			 (X.val[4] != 0x38) || (X.val[5] != 0xc3) ||
			 (X.val[6] != 0x98) || (X.val[7] != 0x77) ||
			 (X.val[8] != 0x14) || (X.val[9] != 0x71) ||
			 (X.val[10] != 0xd9) || (X.val[11] != 0x02) ||
			 (X.val[12] != 0xda) || (X.val[13] != 0x03) ||
			 (X.val[14] != 0xc1) || (X.val[15] != 0x82) ||

			 // 9e6a 5f83 2a93 1d14 b70c 69cd bc8d
			 (X.val[16] != 0x9e) || (X.val[17] != 0x6a) ||
			 (X.val[18] != 0x5f) || (X.val[19] != 0x83) ||
			 (X.val[20] != 0x2a) || (X.val[21] != 0x93) ||
			 (X.val[22] != 0x1d) || (X.val[23] != 0x14) ||
			 (X.val[24] != 0xb7) || (X.val[25] != 0x0c) ||
			 (X.val[26] != 0x69) || (X.val[27] != 0xcd) ||
			 (X.val[28] != 0xbc) || (X.val[29] != 0x8d) )
			return 9;
		break;
	case RAN_DES:
		// 004d fe03 24ca 4971 0fff 3bc0 a734 fceb
		if ( (X.val[0] != 0x00) || (X.val[1] != 0x4d) ||
			 (X.val[2] != 0xfe) || (X.val[3] != 0x03) ||
			 (X.val[4] != 0x24) || (X.val[5] != 0xca) ||
			 (X.val[6] != 0x49) || (X.val[7] != 0x71) ||
			 (X.val[8] != 0x0f) || (X.val[9] != 0xff) ||
			 (X.val[10] != 0x3b) || (X.val[11] != 0xc0) ||
			 (X.val[12] != 0xa7) || (X.val[13] != 0x34) ||
			 (X.val[14] != 0xfc) || (X.val[15] != 0xeb) ||

			 // 4c32 0c0d 2cf0 6abc 9380 373f ef6f
			 (X.val[16] != 0x4c) || (X.val[17] != 0x32) ||
			 (X.val[18] != 0x0c) || (X.val[19] != 0x0d) ||
			 (X.val[20] != 0x2c) || (X.val[21] != 0xf0) ||
			 (X.val[22] != 0x6a) || (X.val[23] != 0xbc) ||
			 (X.val[24] != 0x93) || (X.val[25] != 0x80) ||
			 (X.val[26] != 0x37) || (X.val[27] != 0x3f) ||
			 (X.val[28] != 0xef) || (X.val[29] != 0x6f) )
			return 9;
		break;
	default:
		return 20;
	}

	freeMP(&N);
	freeMP(&X);

	/*  K-283 Curve test  */

	allocMP(&N, 283);
	
	// 01ff ffff ffff ffff ffff ffff ffff ffff ffff 
	// e9ae 2ed0 7577 265d ff7f 9445 1e06 1e16 3c61
	N.val[0] = 0x01; N.val[1] = 0xff;
	N.val[2] = 0xff; N.val[3] = 0xff;
	N.val[4] = 0xff; N.val[5] = 0xff;
	N.val[6] = 0xff; N.val[7] = 0xff;
	N.val[8] = 0xff; N.val[9] = 0xff;
	N.val[10] = 0xff; N.val[11] = 0xff;
	N.val[12] = 0xff; N.val[13] = 0xff;
	N.val[14] = 0xff; N.val[15] = 0xff;
	N.val[16] = 0xff; N.val[17] = 0xff;

	N.val[18] = 0xe9; N.val[19] = 0xae;
	N.val[20] = 0x2e; N.val[21] = 0xd0;
	N.val[22] = 0x75; N.val[23] = 0x77;
	N.val[24] = 0x26; N.val[25] = 0x5d;
	N.val[26] = 0xff; N.val[27] = 0x7f;
	N.val[28] = 0x94; N.val[29] = 0x45;
	N.val[30] = 0x1e; N.val[31] = 0x06;
	N.val[32] = 0x1e; N.val[33] = 0x16;
	N.val[34] = 0x3c; N.val[35] = 0x61;
#ifdef GEN_REF_DATA
fprintBstr(fp, "\n[K-283]\nN is ", N.val, N.size);
#endif
		
	/*  Test ECDSA X9.62 RNG - SHA-Gfnc  */
	XKey.val[0] = 0xbd; XKey.val[1] = 0x02;
	XKey.val[2] = 0x9b; XKey.val[3] = 0xbe;
	XKey.val[4] = 0x7f; XKey.val[5] = 0x51;
	XKey.val[6] = 0x96; XKey.val[7] = 0x0b;
	XKey.val[8] = 0xcf; XKey.val[9] = 0x9e;
	XKey.val[10] = 0xdb; XKey.val[11] = 0x2b;
	XKey.val[12] = 0x61; XKey.val[13] = 0xf0;
	XKey.val[14] = 0x6f; XKey.val[15] = 0x0f;
	XKey.val[16] = 0xeb; XKey.val[17] = 0x5a;
	XKey.val[18] = 0x38; XKey.val[19] = 0xb6;
#ifdef GEN_REF_DATA
fprintBstr(fp, "XKey is ", XKey.val, XKey.size);
#endif
	bzero(XSeed.val, XSeed.size);

	ECDSAGenX(XKey, XSeed, 160, N, &X, rngtype);
	
#ifdef GEN_REF_DATA
	if ( rngtype == RAN_SHS )
fprintBstr(fp, "X-SHA is ", X.val, X.size);
	else if ( rngtype == RAN_DES )
fprintBstr(fp, "X-DES is ", X.val, X.size);
	else
		return 20;
fprintBstr(fp, "Updated XKey is ", XKey.val, XKey.size);
#endif

	switch(rngtype) {
	case RAN_SHS:
		// 01ba 372f de1c 0ffc 7b2e 3b49 8b27 701b
		if ( (X.val[0] != 0x01) || (X.val[1] != 0xba) ||
			 (X.val[2] != 0x37) || (X.val[3] != 0x2f) ||
			 (X.val[4] != 0xde) || (X.val[5] != 0x1c) ||
			 (X.val[6] != 0x0f) || (X.val[7] != 0xfc) ||
			 (X.val[8] != 0x7b) || (X.val[9] != 0x2e) ||
			 (X.val[10] != 0x3b) || (X.val[11] != 0x49) ||
			 (X.val[12] != 0x8b) || (X.val[13] != 0x27) ||
			 (X.val[14] != 0x70) || (X.val[15] != 0x1b) ||

			 // 043f d62f ef68 4d18 fde4 548f 5fec c285
			 (X.val[16] != 0x04) || (X.val[17] != 0x3f) ||
			 (X.val[18] != 0xd6) || (X.val[19] != 0x2f) ||
			 (X.val[20] != 0xef) || (X.val[21] != 0x68) ||
			 (X.val[22] != 0x4d) || (X.val[23] != 0x18) ||
			 (X.val[24] != 0xfd) || (X.val[25] != 0xe4) ||
			 (X.val[26] != 0x54) || (X.val[27] != 0x8f) ||
			 (X.val[28] != 0x5f) || (X.val[29] != 0xec) ||
			 (X.val[30] != 0xc2) || (X.val[31] != 0x85) ||

			 // 86f9 ddd7
			 (X.val[32] != 0x86) || (X.val[33] != 0xf9) ||
			 (X.val[34] != 0xdd) || (X.val[35] != 0xd7) )
			return 10;
		break;
	case RAN_DES:
		// 018d 7f7b ea5a 8f4d fe03 36ff 1170 859a
		if ( (X.val[0] != 0x01) || (X.val[1] != 0x8d) ||
			 (X.val[2] != 0x7f) || (X.val[3] != 0x7b) ||
			 (X.val[4] != 0xea) || (X.val[5] != 0x5a) ||
			 (X.val[6] != 0x8f) || (X.val[7] != 0x4d) ||
			 (X.val[8] != 0xfe) || (X.val[9] != 0x03) ||
			 (X.val[10] != 0x36) || (X.val[11] != 0xff) ||
			 (X.val[12] != 0x11) || (X.val[13] != 0x70) ||
			 (X.val[14] != 0x85) || (X.val[15] != 0x9a) ||

			 // 42c1 52fc 2ade 0b4e dc56 c8ab 5967 8346
			 (X.val[16] != 0x42) || (X.val[17] != 0xc1) ||
			 (X.val[18] != 0x52) || (X.val[19] != 0xfc) ||
			 (X.val[20] != 0x2a) || (X.val[21] != 0xde) ||
			 (X.val[22] != 0x0b) || (X.val[23] != 0x4e) ||
			 (X.val[24] != 0xdc) || (X.val[25] != 0x56) ||
			 (X.val[26] != 0xc8) || (X.val[27] != 0xab) ||
			 (X.val[28] != 0x59) || (X.val[29] != 0x67) ||
			 (X.val[30] != 0x83) || (X.val[31] != 0x46) ||

			 // 9d7d d0d9
			 (X.val[32] != 0x9d) || (X.val[33] != 0x7d) ||
			 (X.val[34] != 0xd0) || (X.val[35] != 0xd9) )
			return 10;
		break;
	default:
		return 20;
	}

	freeMP(&X);

	/*  B-283 Curve test  */
	
	// 03ff ffff ffff ffff ffff ffff ffff ffff ffff
	// ef90 3996 60fc 938a 9016 5b04 2a7c efad b307
	N.val[0] = 0x03; N.val[1] = 0xff;
	N.val[2] = 0xff; N.val[3] = 0xff;
	N.val[4] = 0xff; N.val[5] = 0xff;
	N.val[6] = 0xff; N.val[7] = 0xff;
	N.val[8] = 0xff; N.val[9] = 0xff;
	N.val[10] = 0xff; N.val[11] = 0xff;
	N.val[12] = 0xff; N.val[13] = 0xff;
	N.val[14] = 0xff; N.val[15] = 0xff;
	N.val[16] = 0xff; N.val[17] = 0xff;

	N.val[18] = 0xe9; N.val[19] = 0x90;
	N.val[20] = 0x39; N.val[21] = 0x96;
	N.val[22] = 0x60; N.val[23] = 0xfc;
	N.val[24] = 0x93; N.val[25] = 0x8a;
	N.val[26] = 0x90; N.val[27] = 0x16;
	N.val[28] = 0x5b; N.val[29] = 0x04;
	N.val[30] = 0x2a; N.val[31] = 0x7c;
	N.val[32] = 0xef; N.val[33] = 0xad;
	N.val[34] = 0xb3; N.val[35] = 0x07;
#ifdef GEN_REF_DATA
fprintBstr(fp, "\n[B-283]\nN is ", N.val, N.size);
#endif
		
	/*  Test ECDSA X9.62 RNG - SHA-Gfnc  */
	XKey.val[0] = 0xbd; XKey.val[1] = 0x02;
	XKey.val[2] = 0x9b; XKey.val[3] = 0xbe;
	XKey.val[4] = 0x7f; XKey.val[5] = 0x51;
	XKey.val[6] = 0x96; XKey.val[7] = 0x0b;
	XKey.val[8] = 0xcf; XKey.val[9] = 0x9e;
	XKey.val[10] = 0xdb; XKey.val[11] = 0x2b;
	XKey.val[12] = 0x61; XKey.val[13] = 0xf0;
	XKey.val[14] = 0x6f; XKey.val[15] = 0x0f;
	XKey.val[16] = 0xeb; XKey.val[17] = 0x5a;
	XKey.val[18] = 0x38; XKey.val[19] = 0xb6;
#ifdef GEN_REF_DATA
fprintBstr(fp, "XKey is ", XKey.val, XKey.size);
#endif
	bzero(XSeed.val, XSeed.size);

	ECDSAGenX(XKey, XSeed, 160, N, &X, rngtype);
	
#ifdef GEN_REF_DATA
	if ( rngtype == RAN_SHS )
fprintBstr(fp, "X-SHA is ", X.val, X.size);
	else if ( rngtype == RAN_DES )
fprintBstr(fp, "X-DES is ", X.val, X.size);
	else
		return 20;
fprintBstr(fp, "Updated XKey is ", XKey.val, XKey.size);
#endif

	switch(rngtype) {
	case RAN_SHS:
		// 01ba 372f de1c 0ffc 7b2e 3b49 8b26 bc0a
		if ( (X.val[0] != 0x01) || (X.val[1] != 0xba) ||
			 (X.val[2] != 0x37) || (X.val[3] != 0x2f) ||
			 (X.val[4] != 0xde) || (X.val[5] != 0x1c) ||
			 (X.val[6] != 0x0f) || (X.val[7] != 0xfc) ||
			 (X.val[8] != 0x7b) || (X.val[9] != 0x2e) ||
			 (X.val[10] != 0x3b) || (X.val[11] != 0x49) ||
			 (X.val[12] != 0x8b) || (X.val[13] != 0x26) ||
			 (X.val[14] != 0xbc) || (X.val[15] != 0x0a) ||

			 // 9636 93d4 3c68 e7b5 01d5 b29c 007d b68e
			 (X.val[16] != 0x96) || (X.val[17] != 0x36) ||
			 (X.val[18] != 0x93) || (X.val[19] != 0xd4) ||
			 (X.val[20] != 0x3c) || (X.val[21] != 0x68) ||
			 (X.val[22] != 0xe7) || (X.val[23] != 0xb5) ||
			 (X.val[24] != 0x01) || (X.val[25] != 0xd5) ||
			 (X.val[26] != 0xb2) || (X.val[27] != 0x9c) ||
			 (X.val[28] != 0x00) || (X.val[29] != 0x7d) ||
			 (X.val[30] != 0xb6) || (X.val[31] != 0x8e) ||

			 // a03b a0bd
			 (X.val[32] != 0xa0) || (X.val[33] != 0x3b) ||
			 (X.val[34] != 0xa0) || (X.val[35] != 0xbd) )
			return 11;
		break;
	case RAN_DES:
		// 038d 7f7b ea5a 8f4d fe03 36ff 116b 725d
		if ( (X.val[0] != 0x03) || (X.val[1] != 0x8d) ||
			 (X.val[2] != 0x7f) || (X.val[3] != 0x7b) ||
			 (X.val[4] != 0xea) || (X.val[5] != 0x5a) ||
			 (X.val[6] != 0x8f) || (X.val[7] != 0x4d) ||
			 (X.val[8] != 0xfe) || (X.val[9] != 0x03) ||
			 (X.val[10] != 0x36) || (X.val[11] != 0xff) ||
			 (X.val[12] != 0x11) || (X.val[13] != 0x6b) ||
			 (X.val[14] != 0x72) || (X.val[15] != 0x5d) ||

			 // 8243 8695 40f3 26b5 ea25 e916 22e6 25e5
			 (X.val[16] != 0x82) || (X.val[17] != 0x43) ||
			 (X.val[18] != 0x86) || (X.val[19] != 0x95) ||
			 (X.val[20] != 0x40) || (X.val[21] != 0xf3) ||
			 (X.val[22] != 0x26) || (X.val[23] != 0xb5) ||
			 (X.val[24] != 0xea) || (X.val[25] != 0x25) ||
			 (X.val[26] != 0xe9) || (X.val[27] != 0x16) ||
			 (X.val[28] != 0x22) || (X.val[29] != 0xe6) ||
			 (X.val[30] != 0x25) || (X.val[31] != 0xe5) ||

			 // 6e3a 430b
			 (X.val[32] != 0x6e) || (X.val[33] != 0x3a) ||
			 (X.val[34] != 0x43) || (X.val[35] != 0x0b) )
			return 11;
		break;
	default:
		return 20;
	}

	freeMP(&N);
	freeMP(&X);

	/*  K-409 Curve test  */

	allocMP(&N, 409);
	
	// 007f ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff
	N.val[0] = 0x00; N.val[1] = 0x7f;
	N.val[2] = 0xff; N.val[3] = 0xff;
	N.val[4] = 0xff; N.val[5] = 0xff;
	N.val[6] = 0xff; N.val[7] = 0xff;
	N.val[8] = 0xff; N.val[9] = 0xff;
	N.val[10] = 0xff; N.val[11] = 0xff;
	N.val[12] = 0xff; N.val[13] = 0xff;
	N.val[14] = 0xff; N.val[15] = 0xff;
	N.val[16] = 0xff; N.val[17] = 0xff;
	N.val[18] = 0xff; N.val[19] = 0xff;
	N.val[20] = 0xff; N.val[21] = 0xff;
	N.val[22] = 0xff; N.val[23] = 0xff;
	N.val[24] = 0xff; N.val[25] = 0xff;

	// fe5f 83b2 d4ea 2040 0ec4 557d 5ed3 e3e7 ca5b 4b5c 83b8 e01e 5fcf
	N.val[26] = 0xfe; N.val[27] = 0x5f;
	N.val[28] = 0x83; N.val[29] = 0xb2;
	N.val[30] = 0xd4; N.val[31] = 0xea;
	N.val[32] = 0x20; N.val[33] = 0x40;
	N.val[34] = 0x0e; N.val[35] = 0xc4;
	N.val[36] = 0x55; N.val[37] = 0x7d;
	N.val[38] = 0x5e; N.val[39] = 0xd3;
	N.val[40] = 0xe3; N.val[41] = 0xe7;
	N.val[42] = 0xca; N.val[43] = 0x5b;
	N.val[44] = 0x4b; N.val[45] = 0x5c;
	N.val[46] = 0x83; N.val[47] = 0xb8;
	N.val[48] = 0xe0; N.val[49] = 0x1e;
	N.val[50] = 0x5f; N.val[51] = 0xcf;
#ifdef GEN_REF_DATA
fprintBstr(fp, "\n[K-409]\nN is ", N.val, N.size);
#endif
		
	/*  Test ECDSA X9.62 RNG - SHA-Gfnc  */
	XKey.val[0] = 0xbd; XKey.val[1] = 0x02;
	XKey.val[2] = 0x9b; XKey.val[3] = 0xbe;
	XKey.val[4] = 0x7f; XKey.val[5] = 0x51;
	XKey.val[6] = 0x96; XKey.val[7] = 0x0b;
	XKey.val[8] = 0xcf; XKey.val[9] = 0x9e;
	XKey.val[10] = 0xdb; XKey.val[11] = 0x2b;
	XKey.val[12] = 0x61; XKey.val[13] = 0xf0;
	XKey.val[14] = 0x6f; XKey.val[15] = 0x0f;
	XKey.val[16] = 0xeb; XKey.val[17] = 0x5a;
	XKey.val[18] = 0x38; XKey.val[19] = 0xb6;
#ifdef GEN_REF_DATA
fprintBstr(fp, "XKey is ", XKey.val, XKey.size);
#endif
	bzero(XSeed.val, XSeed.size);

	ECDSAGenX(XKey, XSeed, 160, N, &X, rngtype);
	
#ifdef GEN_REF_DATA
	if ( rngtype == RAN_SHS )
fprintBstr(fp, "X-SHA is ", X.val, X.size);
	else if ( rngtype == RAN_DES )
fprintBstr(fp, "X-DES is ", X.val, X.size);
	else
		return 20;
fprintBstr(fp, "Updated XKey is ", XKey.val, XKey.size);
#endif

	switch(rngtype) {
	case RAN_SHS:
		// 001c 0ffc 7b2e 3b49 8b26 0614 3c6c 18ba
		if ( (X.val[0] != 0x00) || (X.val[1] != 0x1c) ||
			 (X.val[2] != 0x0f) || (X.val[3] != 0xfc) ||
			 (X.val[4] != 0x7b) || (X.val[5] != 0x2e) ||
			 (X.val[6] != 0x3b) || (X.val[7] != 0x49) ||
			 (X.val[8] != 0x8b) || (X.val[9] != 0x26) ||
			 (X.val[10] != 0x06) || (X.val[11] != 0x14) ||
			 (X.val[12] != 0x3c) || (X.val[13] != 0x6c) ||
			 (X.val[14] != 0x18) || (X.val[15] != 0xba) ||

			 // cb0f d5e3 81a6 9c6b 86e5 a33d a830 fca8
			 (X.val[16] != 0xcb) || (X.val[17] != 0x0f) ||
			 (X.val[18] != 0xd5) || (X.val[19] != 0xe3) ||
			 (X.val[20] != 0x81) || (X.val[21] != 0xa6) ||
			 (X.val[22] != 0x9c) || (X.val[23] != 0x6b) ||
			 (X.val[24] != 0x86) || (X.val[25] != 0xe5) ||
			 (X.val[26] != 0xa3) || (X.val[27] != 0x3d) ||
			 (X.val[28] != 0xa8) || (X.val[29] != 0x30) ||
			 (X.val[30] != 0xfc) || (X.val[31] != 0xa8) ||

			 // a832 2710 5af2 d26f 1bdb 2e32 8b9a 491b
			 (X.val[32] != 0xa8) || (X.val[33] != 0x32) ||
			 (X.val[34] != 0x27) || (X.val[35] != 0x10) ||
			 (X.val[36] != 0x5a) || (X.val[37] != 0xf2) ||
			 (X.val[38] != 0xd2) || (X.val[39] != 0x6f) ||
			 (X.val[40] != 0x1b) || (X.val[41] != 0xdb) ||
			 (X.val[42] != 0x2e) || (X.val[43] != 0x32) ||
			 (X.val[44] != 0x8b) || (X.val[45] != 0x9a) ||
			 (X.val[46] != 0x49) || (X.val[47] != 0x1b) ||

			 // c27c 166f
			 (X.val[48] != 0xc2) || (X.val[49] != 0x7c) ||
			 (X.val[50] != 0x16) || (X.val[51] != 0x6f) )
			return 12;
		break;
	case RAN_DES:
		// 005a 8f4d fe03 36ff 1166 516e a319 8afc
		if ( (X.val[0] != 0x00) || (X.val[1] != 0x5a) ||
			 (X.val[2] != 0x8f) || (X.val[3] != 0x4d) ||
			 (X.val[4] != 0xfe) || (X.val[5] != 0x03) ||
			 (X.val[6] != 0x36) || (X.val[7] != 0xff) ||
			 (X.val[8] != 0x11) || (X.val[9] != 0x66) ||
			 (X.val[10] != 0x51) || (X.val[11] != 0x6e) ||
			 (X.val[12] != 0xa3) || (X.val[13] != 0x19) ||
			 (X.val[14] != 0x8a) || (X.val[15] != 0xfc) ||

			 // a402 d0a2 64ed 0c73 f864 c06c 4927 4d3e
			 (X.val[16] != 0xa4) || (X.val[17] != 0x02) ||
			 (X.val[18] != 0xd0) || (X.val[19] != 0xa2) ||
			 (X.val[20] != 0x64) || (X.val[21] != 0xed) ||
			 (X.val[22] != 0x0c) || (X.val[23] != 0x73) ||
			 (X.val[24] != 0xf8) || (X.val[25] != 0x64) ||
			 (X.val[26] != 0xc0) || (X.val[27] != 0x6c) ||
			 (X.val[28] != 0x49) || (X.val[29] != 0x27) ||
			 (X.val[30] != 0x4d) || (X.val[31] != 0x3e) ||

			 // 4c9a 9b9a e779 66bb 78b4 42d8 0879 f9bb
			 (X.val[32] != 0x4c) || (X.val[33] != 0x9a) ||
			 (X.val[34] != 0x9b) || (X.val[35] != 0x9a) ||
			 (X.val[36] != 0xe7) || (X.val[37] != 0x79) ||
			 (X.val[38] != 0x66) || (X.val[39] != 0xbb) ||
			 (X.val[40] != 0x78) || (X.val[41] != 0xb4) ||
			 (X.val[42] != 0x42) || (X.val[43] != 0xd8) ||
			 (X.val[44] != 0x08) || (X.val[45] != 0x79) ||
			 (X.val[46] != 0xf9) || (X.val[47] != 0xbb) ||

			 // 1550 0512
			 (X.val[48] != 0x15) || (X.val[49] != 0x50) ||
			 (X.val[50] != 0x05) || (X.val[51] != 0x12) )
			return 12;
		break;
	default:
		return 20;
	}

	freeMP(&X);

	/*  B-409 Curve test  */
	
	// 0100 0000 0000 0000 0000 0000 0000 0000 0000 0000 0000 0000 0000
	N.val[0] = 0x01; N.val[1] = 0x00;
	N.val[2] = 0x00; N.val[3] = 0x00;
	N.val[4] = 0x00; N.val[5] = 0x00;
	N.val[6] = 0x00; N.val[7] = 0x00;
	N.val[8] = 0x00; N.val[9] = 0x00;
	N.val[10] = 0x00; N.val[11] = 0x00;
	N.val[12] = 0x00; N.val[13] = 0x00;
	N.val[14] = 0x00; N.val[15] = 0x00;
	N.val[16] = 0x00; N.val[17] = 0x00;
	N.val[18] = 0x00; N.val[19] = 0x00;
	N.val[20] = 0x00; N.val[21] = 0x00;
	N.val[22] = 0x00; N.val[23] = 0x00;
	N.val[24] = 0x00; N.val[25] = 0x00;

	// 01e2 aad6 a612 f333 07be 5fa4 7c3c 9e05 2f83 8164 cd37 d9a2 1173
	N.val[26] = 0x01; N.val[27] = 0xe2;
	N.val[28] = 0xaa; N.val[29] = 0xd6;
	N.val[30] = 0xa6; N.val[31] = 0x12;
	N.val[32] = 0xf3; N.val[33] = 0x33;
	N.val[34] = 0x07; N.val[35] = 0xbe;
	N.val[36] = 0x5f; N.val[37] = 0xa4;
	N.val[38] = 0x7c; N.val[39] = 0x3c;
	N.val[40] = 0x9e; N.val[41] = 0x05;
	N.val[42] = 0x2f; N.val[43] = 0x83;
	N.val[44] = 0x81; N.val[45] = 0x64;
	N.val[46] = 0xcd; N.val[47] = 0x37;
	N.val[48] = 0xd9; N.val[49] = 0xa2;
	N.val[50] = 0x11; N.val[51] = 0x73;
#ifdef GEN_REF_DATA
fprintBstr(fp, "\n[B-409]\nN is ", N.val, N.size);
#endif
		
	/*  Test ECDSA X9.62 RNG - SHA-Gfnc  */
	XKey.val[0] = 0xbd; XKey.val[1] = 0x02;
	XKey.val[2] = 0x9b; XKey.val[3] = 0xbe;
	XKey.val[4] = 0x7f; XKey.val[5] = 0x51;
	XKey.val[6] = 0x96; XKey.val[7] = 0x0b;
	XKey.val[8] = 0xcf; XKey.val[9] = 0x9e;
	XKey.val[10] = 0xdb; XKey.val[11] = 0x2b;
	XKey.val[12] = 0x61; XKey.val[13] = 0xf0;
	XKey.val[14] = 0x6f; XKey.val[15] = 0x0f;
	XKey.val[16] = 0xeb; XKey.val[17] = 0x5a;
	XKey.val[18] = 0x38; XKey.val[19] = 0xb6;
#ifdef GEN_REF_DATA
fprintBstr(fp, "XKey is ", XKey.val, XKey.size);
#endif
	bzero(XSeed.val, XSeed.size);

	ECDSAGenX(XKey, XSeed, 160, N, &X, rngtype);
	
#ifdef GEN_REF_DATA
	if ( rngtype == RAN_SHS )
fprintBstr(fp, "X-SHA is ", X.val, X.size);
	else if ( rngtype == RAN_DES )
fprintBstr(fp, "X-DES is ", X.val, X.size);
	else
		return 20;
fprintBstr(fp, "Updated XKey is ", XKey.val, XKey.size);
#endif

	switch(rngtype) {
	case RAN_SHS:
		// 001c 0ffc 7b2e 3b49 8b26 0614 3c6c 18ba
		if ( (X.val[0] != 0x00) || (X.val[1] != 0x1c) ||
			 (X.val[2] != 0x0f) || (X.val[3] != 0xfc) ||
			 (X.val[4] != 0x7b) || (X.val[5] != 0x2e) ||
			 (X.val[6] != 0x3b) || (X.val[7] != 0x49) ||
			 (X.val[8] != 0x8b) || (X.val[9] != 0x26) ||
			 (X.val[10] != 0x06) || (X.val[11] != 0x14) ||
			 (X.val[12] != 0x3c) || (X.val[13] != 0x6c) ||
			 (X.val[14] != 0x18) || (X.val[15] != 0xba) ||

			 // cb0f 2f2b e36a 6549 6eb8 7fd4 3f0b ff68
			 (X.val[16] != 0xcb) || (X.val[17] != 0x0f) ||
			 (X.val[18] != 0x2f) || (X.val[19] != 0x2b) ||
			 (X.val[20] != 0xe3) || (X.val[21] != 0x6a) ||
			 (X.val[22] != 0x65) || (X.val[23] != 0x49) ||
			 (X.val[24] != 0x6e) || (X.val[25] != 0xb8) ||
			 (X.val[26] != 0x7f) || (X.val[27] != 0xd4) ||
			 (X.val[28] != 0x3f) || (X.val[29] != 0x0b) ||
			 (X.val[30] != 0xff) || (X.val[31] != 0x68) ||

			 // c2ff 48a6 0d64 32a8 80ef 58d6 baad 6a70
			 (X.val[32] != 0xc2) || (X.val[33] != 0xff) ||
			 (X.val[34] != 0x48) || (X.val[35] != 0xa6) ||
			 (X.val[36] != 0x0d) || (X.val[37] != 0x64) ||
			 (X.val[38] != 0x32) || (X.val[39] != 0xa8) ||
			 (X.val[40] != 0x80) || (X.val[41] != 0xef) ||
			 (X.val[42] != 0x58) || (X.val[43] != 0xd6) ||
			 (X.val[44] != 0xba) || (X.val[45] != 0xad) ||
			 (X.val[46] != 0x6a) || (X.val[47] != 0x70) ||

			 // 119e d4db
			 (X.val[48] != 0x11) || (X.val[49] != 0x9e) ||
			 (X.val[50] != 0xd4) || (X.val[51] != 0xdb) )
			return 13;
		break;
	case RAN_DES:
		// 005a 8f4d fe03 36ff 1166 516e a319 8afc
		if ( (X.val[0] != 0x00) || (X.val[1] != 0x5a) ||
			 (X.val[2] != 0x8f) || (X.val[3] != 0x4d) ||
			 (X.val[4] != 0xfe) || (X.val[5] != 0x03) ||
			 (X.val[6] != 0x36) || (X.val[7] != 0xff) ||
			 (X.val[8] != 0x11) || (X.val[9] != 0x66) ||
			 (X.val[10] != 0x51) || (X.val[11] != 0x6e) ||
			 (X.val[12] != 0xa3) || (X.val[13] != 0x19) ||
			 (X.val[14] != 0x8a) || (X.val[15] != 0xfc) ||

			 // a3fe 1db3 5918 f5ad 2119 dbd5 5b49 7549
			 (X.val[16] != 0xa3) || (X.val[17] != 0xfe) ||
			 (X.val[18] != 0x1d) || (X.val[19] != 0xb3) ||
			 (X.val[20] != 0x59) || (X.val[21] != 0x18) ||
			 (X.val[22] != 0xf5) || (X.val[23] != 0xad) ||
			 (X.val[24] != 0x21) || (X.val[25] != 0x19) ||
			 (X.val[26] != 0xdb) || (X.val[27] != 0xd5) ||
			 (X.val[28] != 0x5b) || (X.val[29] != 0x49) ||
			 (X.val[30] != 0x75) || (X.val[31] != 0x49) ||

			 // b05b b73a f30b 407c 2488 a423 ff14 39a3
			 (X.val[32] != 0xb0) || (X.val[33] != 0x5b) ||
			 (X.val[34] != 0xb7) || (X.val[35] != 0x3a) ||
			 (X.val[36] != 0xf3) || (X.val[37] != 0x0b) ||
			 (X.val[38] != 0x40) || (X.val[39] != 0x7c) ||
			 (X.val[40] != 0x24) || (X.val[41] != 0x88) ||
			 (X.val[42] != 0xa4) || (X.val[43] != 0x23) ||
			 (X.val[44] != 0xff) || (X.val[45] != 0x14) ||
			 (X.val[46] != 0x39) || (X.val[47] != 0xa3) ||

			 // b937 6576
			 (X.val[48] != 0xb9) || (X.val[49] != 0x37) ||
			 (X.val[50] != 0x65) || (X.val[51] != 0x76) )
			return 13;
		break;
	default:
		return 20;
	}

	freeMP(&N);
	freeMP(&X);

	/*  K-571 Curve test  */

	allocMP(&N, 571);
	
	// 0200 0000 0000 0000 0000 0000 0000 0000 0000
	N.val[0] = 0x02; N.val[1] = 0x00;
	N.val[2] = 0x00; N.val[3] = 0x00;
	N.val[4] = 0x00; N.val[5] = 0x00;
	N.val[6] = 0x00; N.val[7] = 0x00;
	N.val[8] = 0x00; N.val[9] = 0x00;
	N.val[10] = 0x00; N.val[11] = 0x00;
	N.val[12] = 0x00; N.val[13] = 0x00;
	N.val[14] = 0x00; N.val[15] = 0x00;
	N.val[16] = 0x00; N.val[17] = 0x00;

	// 0000 0000 0000 0000 0000 0000 0000 0000 0000
	N.val[18] = 0x00; N.val[19] = 0x00;
	N.val[20] = 0x00; N.val[21] = 0x00;
	N.val[22] = 0x00; N.val[23] = 0x00;
	N.val[24] = 0x00; N.val[25] = 0x00;
	N.val[26] = 0x00; N.val[27] = 0x00;
	N.val[28] = 0x00; N.val[29] = 0x00;
	N.val[30] = 0x00; N.val[31] = 0x00;
	N.val[32] = 0x00; N.val[33] = 0x00;
	N.val[34] = 0x00; N.val[35] = 0x00;

	// 1318 50e1 f19a 63e4 b391 a8db 917f 4138 b630
	N.val[36] = 0x13; N.val[37] = 0x18;
	N.val[38] = 0x50; N.val[39] = 0xe1;
	N.val[40] = 0xf1; N.val[41] = 0x9a;
	N.val[42] = 0x63; N.val[43] = 0xe4;
	N.val[44] = 0xb3; N.val[45] = 0x91;
	N.val[46] = 0xa8; N.val[47] = 0xdb;
	N.val[48] = 0x91; N.val[49] = 0x7f;
	N.val[50] = 0x41; N.val[51] = 0x38;
	N.val[52] = 0xb6; N.val[53] = 0x30;
	
	// d84b e5d6 3938 1e91 deb4 5cfe 778f 637c 1001
	N.val[54] = 0xd8; N.val[55] = 0x4b;
	N.val[56] = 0xe5; N.val[57] = 0xd6;
	N.val[58] = 0x39; N.val[59] = 0x38;
	N.val[60] = 0x1e; N.val[61] = 0x91;
	N.val[62] = 0xde; N.val[63] = 0xb4;
	N.val[64] = 0x5c; N.val[65] = 0xfe;
	N.val[66] = 0x77; N.val[67] = 0x8f;
	N.val[68] = 0x63; N.val[69] = 0x7c;
	N.val[70] = 0x10; N.val[71] = 0x01;
#ifdef GEN_REF_DATA
fprintBstr(fp, "\n[K-571]\nN is ", N.val, N.size);
#endif
		
	/*  Test ECDSA X9.62 RNG - SHA-Gfnc  */
	XKey.val[0] = 0xbd; XKey.val[1] = 0x02;
	XKey.val[2] = 0x9b; XKey.val[3] = 0xbe;
	XKey.val[4] = 0x7f; XKey.val[5] = 0x51;
	XKey.val[6] = 0x96; XKey.val[7] = 0x0b;
	XKey.val[8] = 0xcf; XKey.val[9] = 0x9e;
	XKey.val[10] = 0xdb; XKey.val[11] = 0x2b;
	XKey.val[12] = 0x61; XKey.val[13] = 0xf0;
	XKey.val[14] = 0x6f; XKey.val[15] = 0x0f;
	XKey.val[16] = 0xeb; XKey.val[17] = 0x5a;
	XKey.val[18] = 0x38; XKey.val[19] = 0xb6;
#ifdef GEN_REF_DATA
fprintBstr(fp, "XKey is ", XKey.val, XKey.size);
#endif
	bzero(XSeed.val, XSeed.size);

	ECDSAGenX(XKey, XSeed, 160, N, &X, rngtype);
	
#ifdef GEN_REF_DATA
	if ( rngtype == RAN_SHS )
fprintBstr(fp, "X-SHA is ", X.val, X.size);
	else if ( rngtype == RAN_DES )
fprintBstr(fp, "X-DES is ", X.val, X.size);
	else
		return 20;
fprintBstr(fp, "Updated XKey is ", XKey.val, XKey.size);
#endif

	switch(rngtype) {
	case RAN_SHS:
		// 001c 0ffc 7b2e 3b49 8b26 0614 3c6c 18ba
		if ( (X.val[0] != 0x00) || (X.val[1] != 0x1c) ||
			 (X.val[2] != 0x0f) || (X.val[3] != 0xfc) ||
			 (X.val[4] != 0x7b) || (X.val[5] != 0x2e) ||
			 (X.val[6] != 0x3b) || (X.val[7] != 0x49) ||
			 (X.val[8] != 0x8b) || (X.val[9] != 0x26) ||
			 (X.val[10] != 0x06) || (X.val[11] != 0x14) ||
			 (X.val[12] != 0x3c) || (X.val[13] != 0x6c) ||
			 (X.val[14] != 0x18) || (X.val[15] != 0xba) ||

			 // cb0f 6c55 babb 1378 8e20 d736 6d6e 42ff
			 (X.val[16] != 0xcb) || (X.val[17] != 0x0f) ||
			 (X.val[18] != 0x6c) || (X.val[19] != 0x55) ||
			 (X.val[20] != 0xba) || (X.val[21] != 0xbb) ||
			 (X.val[22] != 0x13) || (X.val[23] != 0x78) ||
			 (X.val[24] != 0x8e) || (X.val[25] != 0x20) ||
			 (X.val[26] != 0xd7) || (X.val[27] != 0x36) ||
			 (X.val[28] != 0x6d) || (X.val[29] != 0x6e) ||
			 (X.val[30] != 0x42) || (X.val[31] != 0xff) ||

			 // 11b7 d851 b401 28dc a031 d10d 0db5 c612
			 (X.val[32] != 0x11) || (X.val[33] != 0xb7) ||
			 (X.val[34] != 0xd8) || (X.val[35] != 0x51) ||
			 (X.val[36] != 0xb4) || (X.val[37] != 0x01) ||
			 (X.val[38] != 0x28) || (X.val[39] != 0xdc) ||
			 (X.val[40] != 0xa0) || (X.val[41] != 0x31) ||
			 (X.val[42] != 0xd1) || (X.val[43] != 0x0d) ||
			 (X.val[44] != 0x0d) || (X.val[45] != 0xb5) ||
			 (X.val[46] != 0xc6) || (X.val[47] != 0x12) ||

			 // 3708 7cd9 f701 f6cf e7f2 d2df 7f78 06e2
			 (X.val[48] != 0x37) || (X.val[49] != 0x08) ||
			 (X.val[50] != 0x7c) || (X.val[51] != 0xd9) ||
			 (X.val[52] != 0xf7) || (X.val[53] != 0x01) ||
			 (X.val[54] != 0xf6) || (X.val[55] != 0xcf) ||
			 (X.val[56] != 0xe7) || (X.val[57] != 0xf2) ||
			 (X.val[58] != 0xd2) || (X.val[59] != 0xdf) ||
			 (X.val[60] != 0x7f) || (X.val[61] != 0x78) ||
			 (X.val[62] != 0x06) || (X.val[63] != 0xe2) ||

			 // b043 e4c6 a9f7 fa5c
			 (X.val[64] != 0xb0) || (X.val[65] != 0x43) ||
			 (X.val[66] != 0xe4) || (X.val[67] != 0xc6) ||
			 (X.val[68] != 0xa9) || (X.val[69] != 0xf7) ||
			 (X.val[70] != 0xfa) || (X.val[71] != 0x5c) )
			return 14;
		break;
	case RAN_DES:
		// 005a 8f4d fe03 36ff 1166 516e a319 8afc
		if ( (X.val[0] != 0x00) || (X.val[1] != 0x5a) ||
			 (X.val[2] != 0x8f) || (X.val[3] != 0x4d) ||
			 (X.val[4] != 0xfe) || (X.val[5] != 0x03) ||
			 (X.val[6] != 0x36) || (X.val[7] != 0xff) ||
			 (X.val[8] != 0x11) || (X.val[9] != 0x66) ||
			 (X.val[10] != 0x51) || (X.val[11] != 0x6e) ||
			 (X.val[12] != 0xa3) || (X.val[13] != 0x19) ||
			 (X.val[14] != 0x8a) || (X.val[15] != 0xfc) ||

			 // a3ff d705 20fd 68c3 2abd fe3f 083e 879a
			 (X.val[16] != 0xa3) || (X.val[17] != 0xff) ||
			 (X.val[18] != 0xd7) || (X.val[19] != 0x05) ||
			 (X.val[20] != 0x20) || (X.val[21] != 0xfd) ||
			 (X.val[22] != 0x68) || (X.val[23] != 0xc3) ||
			 (X.val[24] != 0x2a) || (X.val[25] != 0xbd) ||
			 (X.val[26] != 0xfe) || (X.val[27] != 0x3f) ||
			 (X.val[28] != 0x08) || (X.val[29] != 0x3e) ||
			 (X.val[30] != 0x87) || (X.val[31] != 0x9a) ||

			 // 1cb0 8292 639a 231f 693d b410 4a1b b3f0
			 (X.val[32] != 0x1c) || (X.val[33] != 0xb0) ||
			 (X.val[34] != 0x82) || (X.val[35] != 0x92) ||
			 (X.val[36] != 0x63) || (X.val[37] != 0x9a) ||
			 (X.val[38] != 0x23) || (X.val[39] != 0x1f) ||
			 (X.val[40] != 0x69) || (X.val[41] != 0x3d) ||
			 (X.val[42] != 0xb4) || (X.val[43] != 0x10) ||
			 (X.val[44] != 0x4a) || (X.val[45] != 0x1b) ||
			 (X.val[46] != 0xb3) || (X.val[47] != 0xf0) ||

			 // 8761 57ac 2a69 8fa6 b121 da8c cd36 6061
			 (X.val[48] != 0x87) || (X.val[49] != 0x61) ||
			 (X.val[50] != 0x57) || (X.val[51] != 0xac) ||
			 (X.val[52] != 0x2a) || (X.val[53] != 0x69) ||
			 (X.val[54] != 0x8f) || (X.val[55] != 0xa6) ||
			 (X.val[56] != 0xb1) || (X.val[57] != 0x21) ||
			 (X.val[58] != 0xda) || (X.val[59] != 0x8c) ||
			 (X.val[60] != 0xcd) || (X.val[61] != 0x36) ||
			 (X.val[62] != 0x60) || (X.val[63] != 0x61) ||

			 // 3258 29cc d029 3aa9
			 (X.val[64] != 0x32) || (X.val[65] != 0x58) ||
			 (X.val[66] != 0x29) || (X.val[67] != 0xcc) ||
			 (X.val[68] != 0xd0) || (X.val[69] != 0x29) ||
			 (X.val[70] != 0x3a) || (X.val[71] != 0xa9) )
			return 14;
		break;
	default:
		return 20;
	}

	freeMP(&X);

	/*  B-571 Curve test  */
	
	// 03ff ffff ffff ffff ffff ffff ffff ffff ffff
	N.val[0] = 0x03; N.val[1] = 0xff;
	N.val[2] = 0xff; N.val[3] = 0xff;
	N.val[4] = 0xff; N.val[5] = 0xff;
	N.val[6] = 0xff; N.val[7] = 0xff;
	N.val[8] = 0xff; N.val[9] = 0xff;
	N.val[10] = 0xff; N.val[11] = 0xff;
	N.val[12] = 0xff; N.val[13] = 0xff;
	N.val[14] = 0xff; N.val[15] = 0xff;
	N.val[16] = 0xff; N.val[17] = 0xff;

	// ffff ffff ffff ffff ffff ffff ffff ffff ffff
	N.val[18] = 0xff; N.val[19] = 0xff;
	N.val[20] = 0xff; N.val[21] = 0xff;
	N.val[22] = 0xff; N.val[23] = 0xff;
	N.val[24] = 0xff; N.val[25] = 0xff;
	N.val[26] = 0xff; N.val[27] = 0xff;
	N.val[28] = 0xff; N.val[29] = 0xff;
	N.val[30] = 0xff; N.val[31] = 0xff;
	N.val[32] = 0xff; N.val[33] = 0xff;
	N.val[34] = 0xff; N.val[35] = 0xff;

	// e661 ce18 ff55 9873 0805 9b18 6823 851e c7dd
	N.val[36] = 0xe6; N.val[37] = 0x61;
	N.val[38] = 0xce; N.val[39] = 0x18;
	N.val[40] = 0xff; N.val[41] = 0x55;
	N.val[42] = 0x98; N.val[43] = 0x73;
	N.val[44] = 0x08; N.val[45] = 0x05;
	N.val[46] = 0x9b; N.val[47] = 0x18;
	N.val[48] = 0x68; N.val[49] = 0x23;
	N.val[50] = 0x85; N.val[51] = 0x1e;
	N.val[52] = 0xc7; N.val[53] = 0xdd;
	
	// 9ca1 161d e93d 5174 d66e 8382 e9bb 2fe8 4e47
	N.val[54] = 0x9c; N.val[55] = 0xa1;
	N.val[56] = 0x16; N.val[57] = 0x1d;
	N.val[58] = 0xe9; N.val[59] = 0x3d;
	N.val[60] = 0x51; N.val[61] = 0x74;
	N.val[62] = 0xd6; N.val[63] = 0x6e;
	N.val[64] = 0x83; N.val[65] = 0x82;
	N.val[66] = 0xe9; N.val[67] = 0xbb;
	N.val[68] = 0x2f; N.val[69] = 0xe8;
	N.val[70] = 0x4e; N.val[71] = 0x47;
#ifdef GEN_REF_DATA
fprintBstr(fp, "\n[B-571]\nN is ", N.val, N.size);
#endif
		
	/*  Test ECDSA X9.62 RNG - SHA-Gfnc  */
	XKey.val[0] = 0xbd; XKey.val[1] = 0x02;
	XKey.val[2] = 0x9b; XKey.val[3] = 0xbe;
	XKey.val[4] = 0x7f; XKey.val[5] = 0x51;
	XKey.val[6] = 0x96; XKey.val[7] = 0x0b;
	XKey.val[8] = 0xcf; XKey.val[9] = 0x9e;
	XKey.val[10] = 0xdb; XKey.val[11] = 0x2b;
	XKey.val[12] = 0x61; XKey.val[13] = 0xf0;
	XKey.val[14] = 0x6f; XKey.val[15] = 0x0f;
	XKey.val[16] = 0xeb; XKey.val[17] = 0x5a;
	XKey.val[18] = 0x38; XKey.val[19] = 0xb6;
#ifdef GEN_REF_DATA
fprintBstr(fp, "XKey is ", XKey.val, XKey.size);
#endif
	bzero(XSeed.val, XSeed.size);

	ECDSAGenX(XKey, XSeed, 160, N, &X, rngtype);
	
#ifdef GEN_REF_DATA
	if ( rngtype == RAN_SHS )
fprintBstr(fp, "X-SHA is ", X.val, X.size);
	else if ( rngtype == RAN_DES )
fprintBstr(fp, "X-DES is ", X.val, X.size);
	else
		return 20;
fprintBstr(fp, "Updated XKey is ", XKey.val, XKey.size);
#endif

	switch(rngtype) {
	case RAN_SHS:
		// 021c 0ffc 7b2e 3b49 8b26 0614 3c6c 18ba
		if ( (X.val[0] != 0x02) || (X.val[1] != 0x1c) ||
			 (X.val[2] != 0x0f) || (X.val[3] != 0xfc) ||
			 (X.val[4] != 0x7b) || (X.val[5] != 0x2e) ||
			 (X.val[6] != 0x3b) || (X.val[7] != 0x49) ||
			 (X.val[8] != 0x8b) || (X.val[9] != 0x26) ||
			 (X.val[10] != 0x06) || (X.val[11] != 0x14) ||
			 (X.val[12] != 0x3c) || (X.val[13] != 0x6c) ||
			 (X.val[14] != 0x18) || (X.val[15] != 0xba) ||

			 // cb0f 6c55 babb 1378 8e20 d738 72ea a906
			 (X.val[16] != 0xcb) || (X.val[17] != 0x0f) ||
			 (X.val[18] != 0x6c) || (X.val[19] != 0x55) ||
			 (X.val[20] != 0xba) || (X.val[21] != 0xbb) ||
			 (X.val[22] != 0x13) || (X.val[23] != 0x78) ||
			 (X.val[24] != 0x8e) || (X.val[25] != 0x20) ||
			 (X.val[26] != 0xd7) || (X.val[27] != 0x38) ||
			 (X.val[28] != 0x72) || (X.val[29] != 0xea) ||
			 (X.val[30] != 0xa9) || (X.val[31] != 0x06) ||

			 // 9870 dbbe 915d d82f 86f2 871d 0f08 edff
			 (X.val[32] != 0x98) || (X.val[33] != 0x70) ||
			 (X.val[34] != 0xdb) || (X.val[35] != 0xbe) ||
			 (X.val[36] != 0x91) || (X.val[37] != 0x5d) ||
			 (X.val[38] != 0xd8) || (X.val[39] != 0x2f) ||
			 (X.val[40] != 0x86) || (X.val[41] != 0xf2) ||
			 (X.val[42] != 0x87) || (X.val[43] != 0x1d) ||
			 (X.val[44] != 0x0f) || (X.val[45] != 0x08) ||
			 (X.val[46] != 0xed) || (X.val[47] != 0xff) ||

			 // edff 0856 6f29 3949 8359 24b1 bcac a9d8
			 (X.val[48] != 0xed) || (X.val[49] != 0xff) ||
			 (X.val[50] != 0x08) || (X.val[51] != 0x56) ||
			 (X.val[52] != 0x6f) || (X.val[53] != 0x29) ||
			 (X.val[54] != 0x39) || (X.val[55] != 0x49) ||
			 (X.val[56] != 0x83) || (X.val[57] != 0x59) ||
			 (X.val[58] != 0x24) || (X.val[59] != 0xb1) ||
			 (X.val[60] != 0xbc) || (X.val[61] != 0xac) ||
			 (X.val[62] != 0xa9) || (X.val[63] != 0xd8) ||

			 // 43ad fef0 c877 e2d2
			 (X.val[64] != 0x43) || (X.val[65] != 0xad) ||
			 (X.val[66] != 0xfe) || (X.val[67] != 0xf0) ||
			 (X.val[68] != 0xc8) || (X.val[69] != 0x77) ||
			 (X.val[70] != 0xe2) || (X.val[71] != 0xd2) )
			return 15;
		break;
	case RAN_DES:
		// 025a 8f4d fe03 36ff 1166 516e a319 8afc
		if ( (X.val[0] != 0x02) || (X.val[1] != 0x5a) ||
			 (X.val[2] != 0x8f) || (X.val[3] != 0x4d) ||
			 (X.val[4] != 0xfe) || (X.val[5] != 0x03) ||
			 (X.val[6] != 0x36) || (X.val[7] != 0xff) ||
			 (X.val[8] != 0x11) || (X.val[9] != 0x66) ||
			 (X.val[10] != 0x51) || (X.val[11] != 0x6e) ||
			 (X.val[12] != 0xa3) || (X.val[13] != 0x19) ||
			 (X.val[14] != 0x8a) || (X.val[15] != 0xfc) ||

			 // a3ff d705 20fd 68c3 2abd fe4d 9e1e 614a
			 (X.val[16] != 0xa3) || (X.val[17] != 0xff) ||
			 (X.val[18] != 0xd7) || (X.val[19] != 0x05) ||
			 (X.val[20] != 0x20) || (X.val[21] != 0xfd) ||
			 (X.val[22] != 0x68) || (X.val[23] != 0xc3) ||
			 (X.val[24] != 0x2a) || (X.val[25] != 0xbd) ||
			 (X.val[26] != 0xfe) || (X.val[27] != 0x4d) ||
			 (X.val[28] != 0x9e) || (X.val[29] != 0x1e) ||
			 (X.val[30] != 0x61) || (X.val[31] != 0x4a) ||

			 // e01a 843b ca64 203d ffbe eda4 322c 1cd6
			 (X.val[32] != 0xe0) || (X.val[33] != 0x1a) ||
			 (X.val[34] != 0x84) || (X.val[35] != 0x3b) ||
			 (X.val[36] != 0xca) || (X.val[37] != 0x64) ||
			 (X.val[38] != 0x20) || (X.val[39] != 0x3d) ||
			 (X.val[40] != 0xff) || (X.val[41] != 0xbe) ||
			 (X.val[42] != 0xed) || (X.val[43] != 0xa4) ||
			 (X.val[44] != 0x32) || (X.val[45] != 0x2c) ||
			 (X.val[46] != 0x1c) || (X.val[47] != 0xd6) ||

			 // 7b0d e3fc ce28 8864 5ed0 92d1 657d 2c63
			 (X.val[48] != 0x7b) || (X.val[49] != 0x0d) ||
			 (X.val[50] != 0xe3) || (X.val[51] != 0xfc) ||
			 (X.val[52] != 0xce) || (X.val[53] != 0x28) ||
			 (X.val[54] != 0x88) || (X.val[55] != 0x64) ||
			 (X.val[56] != 0x5e) || (X.val[57] != 0xd0) ||
			 (X.val[58] != 0x92) || (X.val[59] != 0xd1) ||
			 (X.val[60] != 0x65) || (X.val[61] != 0x7d) ||
			 (X.val[62] != 0x2c) || (X.val[63] != 0x63) ||

			 // e6b9 6946 32fd 664d
			 (X.val[64] != 0xe6) || (X.val[65] != 0xb9) ||
			 (X.val[66] != 0x69) || (X.val[67] != 0x46) ||
			 (X.val[68] != 0x32) || (X.val[69] != 0xfd) ||
			 (X.val[70] != 0x66) || (X.val[71] != 0x4d) )
			return 15;
		break;
	default:
		return 20;
	}

	freeMP(&N);
	freeMP(&X);
	freeMP(&XKey);
	freeMP(&XSeed);

#ifdef GEN_REF_DATA
	fclose(fp);
#endif

	return 0;
}

/***************/
/*  ANSI X9.31 */
/***************/
void ANSIX931Init(TDESCTX *context, BYTE *Key1, BYTE *Key2)
{
	BYTE	Key3[8];

	bcopy(Key1, Key3, 8);
	TDESInit (context, Key1, Key2, Key3, NULL, TDES_ECB, DES_ENCIPHER);
}

void ANSIX931GenR(TDESCTX *context, BYTE *V, BYTE *DT, BYTE *R)
{
	BYTE	I[8], I2[8];
	int		i;
//	FILE	*fp;
//	fp = fopen("atest2.txt", "a");

	TDES_Core(context, DT, I);
	for ( i=0; i<8; i++ )
		I2[i] = I[i] ^ V[i];

	TDES_Core(context, I2, R);

	for ( i=0; i<8; i++ )
		I2[i] = R[i] ^ I[i];

	TDES_Core(context, I2, V);

	for ( i=7; i>=0; i-- ) {
		if ( DT[i] == 0xff )
			DT[i] = 0;
		else {
			DT[i]++;
			break;
		}
	}
/*	for ( i=0; i<8; i++ ) {
		fprintf(fp, "%c", (R[i]&0x80) ? '1' : '0');
		fprintf(fp, "%c", (R[i]&0x40) ? '1' : '0');
		fprintf(fp, "%c", (R[i]&0x20) ? '1' : '0');
		fprintf(fp, "%c", (R[i]&0x10) ? '1' : '0');
		fprintf(fp, "%c", (R[i]&0x08) ? '1' : '0');
		fprintf(fp, "%c", (R[i]&0x04) ? '1' : '0');
		fprintf(fp, "%c", (R[i]&0x02) ? '1' : '0');
		fprintf(fp, "%c ", (R[i]&0x01) ? '1' : '0');
	}
	fprintf(fp, "\n");
	fclose(fp);
*/
}

int ANSIX931RngTest()
{
	BYTE	Key1[8], Key2[8], DT[8], V[8], R[8];
	TDESCTX	context;
	/*
	FILE	*fp;
	fp = fopen("atest2.txt", "w");
	*/


	// Key1 = 75c7 1ae5 a11a 232c
	Key1[0] = 0x75; Key1[1] = 0xc7;
	Key1[2] = 0x1a; Key1[3] = 0xe5;
	Key1[4] = 0xa1; Key1[5] = 0x1a;
	Key1[6] = 0x23; Key1[7] = 0x2c;

	// Key2 = 4025 6dcd 94f7 67b0
	Key2[0] = 0x40; Key2[1] = 0x25;
	Key2[2] = 0x6d; Key2[3] = 0xcd;
	Key2[4] = 0x94; Key2[5] = 0xf7;
	Key2[6] = 0x67; Key2[7] = 0xb0;

	// DT = c89a 1d88 8ed1 2f3c
	DT[0] = 0xc8; DT[1] = 0x9a;
	DT[2] = 0x1d; DT[3] = 0x88;
	DT[4] = 0x8e; DT[5] = 0xd1;
	DT[6] = 0x2f; DT[7] = 0x3c;

	// V = 8000 0000 0000 0000
	V[0] = 0x80; V[1] = 0x00;
	V[2] = 0x00; V[3] = 0x00;
	V[4] = 0x00; V[5] = 0x00;
	V[6] = 0x00; V[7] = 0x00;
	
	ANSIX931Init(&context, Key1, Key2);
	ANSIX931GenR(&context, V, DT, R);
	
	// R = 944d c721 0d6d 7fd7
	if ( (R[0] != 0x94) || (R[1] != 0x4d) ||
		 (R[2] != 0xc7) || (R[3] != 0x21) ||
		 (R[4] != 0x0d) || (R[5] != 0x6d) ||
		 (R[6] != 0x7f) || (R[7] != 0xd7) )
		return 1;

	return 0;
}

/*  I'm not currently calling this routine.  Instead I call the individual
    routines as they are needed.  That is, if you only do DSA - only the
	DSA test is called (see FIPS186_gen in rngtab.cpp).  */
int RNG_Test()
{
	int		ret_val;

	if ( ret_val = DSARngTest() )
		return ret_val + 100;

	if ( ret_val = ECDSARngTest(RAN_SHS) )
		return ret_val + 200;

	if ( ret_val = ECDSARngTest(RAN_DES) )
		return ret_val + 300;

	if ( ret_val = ANSIX931RngTest() )
		return ret_val + 400;

	return 0;
}